# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                        initial documentation
#        :author: Francois Gallard, Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
The Individual Discipline Feasible formulation
**********************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import str, super

from future import standard_library
from numpy import abs as np_abs
from numpy import concatenate, eye, ones_like, zeros

from gemseo import LOGGER
from gemseo.core.chain import MDOParallelChain
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.core.execution_sequence import ExecutionSequenceFactory
from gemseo.core.formulation import MDOFormulation
from gemseo.core.function import MDOFunction
from gemseo.mda.mda_chain import MDAChain

standard_library.install_aliases()


class IDF(MDOFormulation):
    """The Individual Discipline Feasible formulation.

    This formulation draws an optimization architecture where the coupling
    variables of strongly coupled disciplines is made consistent by adding
    equality constraints on the coupling variables at top level, the
    optimization problem w.r.t. local, global design variables and coupling
    variables is made at the top level. Disciplinary analysis is made at a each
    optimization iteration. Multidisciplinary analysis is made at the optimum.
    """

    def __init__(
        self,
        disciplines,
        objective_name,
        design_space,
        maximize_objective=False,
        normalize_constraints=True,
        parallel_exec=False,
        use_threading=True,
        start_at_equilibrium=False,
    ):
        """
        Constructor, initializes the objective functions and constraints

        :param disciplines: the disciplines list.
        :type disciplines: list(MDODiscipline)
        :param objective_name: the objective function data name.
        :type objective_name: str
        :param design_space: the design space.
        :type design_space: DesignSpace
        :param maximize_objective: if True, the objective function
            is maximized, by default, a minimization is performed.
        :type maximize_objective: bool
        :param normalize_constraints: if True, outputs of the coupling
            consistency contraints are scaled.
        :type normalize_constraints: bool
        :param parallel_exec: if True, all constraints and objectives
            are computed in parallel. At every iteration, all disciplines
            are executed in parallel. Otherwise, a separate constraint is
            created for each discipline with couplings.
        :type parallel_exec: bool
        :param use_threading: if True and parallel_exec=True,
             the disciplines are run in parallel using multi-threading,
             if False and parallel_exec=True, multi-processing is used.
        :type use_threading: bool
        :param start_at_equilibrium: if True, an MDA is used to initialize
            coupling variables.
        :type start_at_equilibrium: bool
        """
        super(IDF, self).__init__(
            disciplines,
            objective_name,
            design_space,
            maximize_objective=maximize_objective,
        )
        if parallel_exec:
            self._parallel_exec = MDOParallelChain(
                self.disciplines, use_threading=use_threading
            )
        else:
            self._parallel_exec = None

        self.coupling_structure = MDOCouplingStructure(disciplines)
        self.all_couplings = self.coupling_structure.get_all_couplings()
        self._update_design_space()
        self.normalize_constraints = normalize_constraints
        self._build_constraints()
        self._build_objective_from_disc(objective_name)

        if start_at_equilibrium:
            self._compute_equilibrium()

    def _compute_equilibrium(self):
        """Run an MDA to compute initial target couplings at equilibrium.

        Values at equlibrium are set in the initial design space.
        """
        current_x = self.design_space.get_current_x_dict()
        # run MDA to initialize target coupling variables
        mda = MDAChain(self.disciplines)
        res = mda.execute(current_x)

        for name in self.all_couplings:
            value = res[name]
            LOGGER.info(
                "IDF: changing the initial value of %s " "from %s to %s (equilibrium)",
                name,
                str(current_x[name]),
                str(value),
            )
            self.design_space.set_current_variable(name, value)

        # Reset the number of evaluations
        for disc in self.disciplines:
            disc.n_calls = 0
            disc.n_calls_linearize = 0

    def _update_design_space(self):
        """Updates the design space with the required variables"""
        strong_couplings = set(self.all_couplings)
        variables_names = set(self.opt_problem.design_space.variables_names)
        if not strong_couplings.issubset(variables_names):
            missing = strong_couplings - variables_names
            raise ValueError(
                "IDF formulation needs coupling variables"
                + " as design variables, missing variables : "
                + str(missing)
            )
        self._set_defaultinputs_from_ds()

    def get_top_level_disc(self):
        # All functions and constraints are built from the top level disc
        # If we are in parallel mode: return the parallel execution
        if self._parallel_exec is not None:
            return [self._parallel_exec]
        # Otherwise the disciplines are top level
        return self.disciplines

    def _get_normalization_factor(self, output_couplings):
        """Returns [abs(ub-lb)] for all outputs_couplings.

        :param output_couplings: names of variables for normalization
        """
        norm_fact = []
        for output in output_couplings:
            u_b = self.design_space.get_upper_bound(output)
            l_b = self.design_space.get_lower_bound(output)
            norm_fact.append(np_abs(u_b - l_b))
        return concatenate(norm_fact)

    def _generate_consistency_cstr(self, output_couplings):
        """Generates the consistency constraints for a discipline.

        :param output_coupl: the output coupling variables
            of the discipline
        :param output_couplings: returns: mdo_c :
            function pointer to consistency constraints
        :returns: mdo_c : function pointer to consistency constraints

        """
        coupl_func = self._get_function_from(output_couplings)
        dv_names_of_disc = coupl_func.args

        if self.normalize_constraints:
            norm_fact = self._get_normalization_factor(output_couplings)
        else:
            norm_fact = 1.0

        def coupl_min_x(x_vec):
            """Function to compute consistency constraints.

            :param x: design variable vector
            :param x_vec: returns: value of consistency constraints
                (=0 if disciplines are at equilibrium)
            :returns: value of consistency constraints
                (=0 if disciplines are at equilibrium)
            """
            x_sw = self.mask_x_swap_order(output_couplings, x_vec)
            coupl = coupl_func(x_vec)
            if self.normalize_constraints:
                return (coupl - x_sw) / norm_fact
            return coupl - x_sw

        def coupl_min_x_jac(x_vec):
            """Function to compute consistency constraints gradient.

            :param x: design variable vector
            :param x_vec: returns: gradient of consistency constraints
            :returns: gradient of consistency constraints
            """
            coupl_jac = coupl_func.jac(x_vec)  # pylint: disable=E1102

            if len(coupl_jac.shape) > 1:
                # IN this case it is harder since a block diagonal
                # matrix with -Id should be placed for each output
                # coupling, at the right place
                n_outs = coupl_jac.shape[0]
                x_jac_2d = zeros((n_outs, len(x_vec)), dtype=x_vec.dtype)
                x_names = self.get_optim_variables_names()
                o_min = 0
                o_max = 0
                for out in output_couplings:
                    # self._reference_input_data[out].size
                    o_len = self._get_dv_length(out)
                    i_min = 0
                    i_max = 0
                    o_max += o_len
                    for x_i in x_names:
                        # self._reference_input_data[x_i].size
                        x_len = self._get_dv_length(x_i)
                        i_max += x_len
                        if x_i == out:
                            x_jac_2d[o_min:o_max, i_min:i_max] = eye(x_len)
                        i_min = i_max
                    o_min = o_max
                x_jac = x_jac_2d
            else:
                # This is surprising but there is a duality between the masking
                # operation in the function inputs and the unmasking of its
                # outputs
                x_jac = self.unmask_x_swap_order(output_couplings, ones_like(x_vec))
            if self.normalize_constraints:
                return (coupl_jac - x_jac) / norm_fact[:, None]
            return coupl_jac - x_jac

        expr = ""
        for out_c in output_couplings:
            expr += out_c + "(" + ", ".join(dv_names_of_disc) + ") - "
            expr += str(out_c) + "" + "\n"

        name = coupl_func.name
        return MDOFunction(
            coupl_min_x,
            name,
            args=dv_names_of_disc,
            expr=expr,
            jac=coupl_min_x_jac,
            outvars=coupl_func.outvars,
            f_type=MDOFunction.TYPE_EQ,
        )

    def _build_constraints(self):
        """In IDF formulation, the consistency constraints are "y - y_copy = 0"
        fills self._eq_constraints attributes
        """
        # Building constraints per generator couplings
        for discipline in self.disciplines:
            couplings = self.coupling_structure.output_couplings(
                discipline, strong=False
            )
            if couplings:
                cstr = self._generate_consistency_cstr(couplings)
                self.opt_problem.add_eq_constraint(cstr)

    def get_expected_workflow(self):
        """
        Returns the expected execution sequence,
        used for xdsm representation
        See MDOFormulation.get_expected_workflow
        """
        return ExecutionSequenceFactory.parallel(self.disciplines)

    def get_expected_dataflow(self):
        """
        Returns the expected data exchange sequence,
        used for xdsm representation
        See MDOFormulation.get_expected_dataflow
        """
        return []



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Damien Guenot
#       :author: Francois Gallard, Charlie Vanaret, Benoit Pauwels
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
    Optimization problem
    ====================

    The :class:`.OptimizationProblem` class is used to define
    the optimization problem from a :class:`.DesignSpace` defining:

    - Initial guess :math:`x_0`
    - Bounds :math:`l_b \\leq x \\leq u_b`

    (Possible vector) objective function is defined as a :class:`.MDOFunction`
    and set using the :code:`objective` attribute.
    If the optimization problem looks for the maximum of this objective
    function, the :meth:`.OptimizationProblem.change_objective_sign` changes
    the objective function sign because the optimization drivers seek to
    minimize this objective function.

    Equality and inequality constraints are also :class:`.MDOFunction` s
    provided to the :class:`.OptimizationProblem` by means of its
    :meth:`.OptimizationProblem.add_constraint` method.

    The :class:`.OptimizationProblem` allows to evaluate the different
    functions for a given design parameters vector
    (see :meth:`.OptimizationProblem.evaluate_functions`). Note that
    this evaluation step relies on an automated scaling of function wrt bounds
    so that optimizers and DOE algorithms work with scaled inputs
    between 0 and 1 for all variables.
    The :class:`.OptimizationProblem`  has also a :class:`.Database` storing
    the calls to all functions so that no function is called twice
    with the same inputs. Concerning the derivatives computation,
    the :class:`.OptimizationProblem` automates the generation of finite
    differences or complex step wrappers on functions,
    when analytical gradient is not available.

    Lastly, various getters and setters are available, as well as methods
    to export the :class:`.Database` to an HDF file or to a :class:`.Dataset`
    for future post-processing.
"""

from __future__ import absolute_import, division, unicode_literals

from functools import reduce
from numbers import Number

import h5py
from future import standard_library
from numpy import abs as np_abs
from numpy import all as np_all
from numpy import any as np_any
from numpy import argmin, array, concatenate, inf
from numpy import isnan as np_isnan
from numpy import issubdtype, ndarray
from numpy import number as np_number
from numpy import string_ as np_string
from numpy import where
from numpy.linalg import norm
from six import string_types

from gemseo.algos.database import Database
from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_result import OptimizationResult
from gemseo.algos.stop_criteria import DesvarIsNan, FunctionIsNan
from gemseo.core.dataset import Dataset
from gemseo.core.function import MDOFunction
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.derivatives_approx import ComplexStep, FirstOrderFD
from gemseo.utils.py23_compat import PY3, string_array

standard_library.install_aliases()
from gemseo import LOGGER


class OptimizationProblem(object):
    """An optimization problem class to store:

    - A (possibly vector) objective function
    - Constraints, equality and inequality
    - Initial guess x_0
    - Bounds l_bounds<= x <= u_bounds
    - The database of calls to all functions so that no function is called
      twice with the same inputs

    It also has an automated scaling of function wrt bounds
    so that optimizers and DOE algorithms work with scaled inputs
    between 0 and 1 for all variables.

    It automates the generation of finite differences or complex
    step wrappers on functions, when analytical gradient
    is not available
    """

    LINEAR_PB = "linear"
    NON_LINEAR_PB = "non-linear"
    AVAILABLE_PB_TYPES = [LINEAR_PB, NON_LINEAR_PB]

    USER_GRAD = "user"
    COMPLEX_STEP = "complex_step"
    FINITE_DIFFERENCES = "finite_differences"
    NO_DERIVATIVES = "no_derivatives"
    DIFFERENTIATION_METHODS = [
        USER_GRAD,
        COMPLEX_STEP,
        FINITE_DIFFERENCES,
        NO_DERIVATIVES,
    ]
    DESIGN_VAR_NAMES = "x_names"
    DESIGN_VAR_SIZE = "x_size"
    DESIGN_SPACE_ATTRS = ["u_bounds", "l_bounds", "x_0", DESIGN_VAR_NAMES, "dimension"]
    FUNCTIONS_ATTRS = ["objective", "constraints"]
    OPTIM_DESCRIPTION = [
        "minimize_objective",
        "fd_step",
        "differentiation_method",
        "pb_type",
        "ineq_tolerance",
        "eq_tolerance",
    ]

    OPT_DESCR_GROUP = "opt_description"
    DESIGN_SPACE_GROUP = "design_space"
    OBJECTIVE_GROUP = "objective"
    SOLUTION_GROUP = "solution"
    CONSTRAINTS_GROUP = "constraints"

    HDF5_FORMAT = "hdf5"
    GGOBI_FORMAT = "ggobi"

    def __init__(
        self,
        design_space,
        pb_type=NON_LINEAR_PB,
        input_database=None,
        differentiation_method=USER_GRAD,
        fd_step=1e-7,
    ):
        """
        Constructor for the optimization problem

        :param pb_type: the type of optimization problem among
               OptimizationProblem.AVAILABLE_PB_TYPES
        :param input_database: a file to eventually load the database
        :param differentiation_method: the default differentiation method
               for the functions of the optimization problem
        :param fd_step: finite differences or complex step step
        """
        self.objective = None
        self.nonproc_objective = None
        self.constraints = []
        self.nonproc_constraints = []
        self.observables = []
        self.new_iter_observables = []
        self.nonproc_observables = []
        self.nonproc_new_iter_observables = []
        self.minimize_objective = True
        self.fd_step = fd_step
        self.differentiation_method = differentiation_method
        self.pb_type = pb_type
        self.ineq_tolerance = 1e-4
        self.eq_tolerance = 1e-2
        self.__functions_are_preprocessed = False
        self.database = Database(input_hdf_file=input_database)
        self.solution = None
        self.design_space = design_space
        self.__x0 = None
        self.stop_if_nan = True
        self.__store_listeners = []
        self.__newiter_listeners = []
        self.preprocess_options = {}

    def add_constraint(self, cstr_func, value=None, cstr_type=None, positive=False):
        """Add constraints (equality and inequality) from MDOFunction

        :param cstr_func: constraints as an MDOFunction
        :param value: the target value for the constraint
            by default cstr(x)<= 0 or cstr(x)= 0 otherwise cstr(x)<=value
        :param cstr_type: constraint type (equality or inequality)
            (Default value = None)
        :param positive: positive/negative inequality constraint
            (Default value = False)
        """
        self.check_format(cstr_func)
        if value is not None:
            cstr_func = cstr_func.offset(-value)
        if positive:
            cstr_func = -cstr_func  # Operator overloading in MDOFunction
        if cstr_type is not None:
            cstr_func.f_type = cstr_type
        else:
            if not cstr_func.is_constraint():
                msg = "Constraint type must be provided, either in the "
                msg += "function attributes or to the add_constraint method"
                raise ValueError(msg)
        self.constraints.append(cstr_func)

    def add_eq_constraint(self, cstr_func, value=None):
        """Add equality constraints to the optimization problem

        :param cstr_func: MDOFunction constraints
        :param value: the target value for the constraint
            by default, cstr(x)=0 otherwise cstr(x)=value
        """
        self.add_constraint(cstr_func, value, cstr_type=MDOFunction.TYPE_EQ)

    def add_ineq_constraint(self, cstr_func, value=None, positive=False):
        """Add inequality constraints to the optimization problem

        :param cstr_func: MDOFunction constraints
        :param value: the target value for the constraint
            by default, cstr(x)<= 0 otherwise cstr(x)<=value
        :param positive: if True, the constraint should be
            cstr(x)>= value, by default  cstr(x)<= value
        """
        self.add_constraint(
            cstr_func, value, cstr_type=MDOFunction.TYPE_INEQ, positive=positive
        )

    def add_observable(self, obs_func, new_iter=True):
        """Adds observable as an MDOFunction.

        :param obs_func: observable as an MDOFunction
        :type obs_func: MDOFunction
        :param new_iter: if True, the observable will be called at each new
            iterate
        :type new_iter: bool
        """
        self.check_format(obs_func)
        obs_func.f_type = MDOFunction.TYPE_OBS
        self.observables.append(obs_func)
        if new_iter:
            self.new_iter_observables.append(obs_func)

    def get_eq_constraints(self):
        """Accessor for all equality constraints


        :returns: a list of equality constraints
        """

        def filter_eq(cstr):
            """A filter for equality constraints

            :param cstr: constraint function
            :returns: True if the function is an equality constraint
            """
            return cstr.f_type == MDOFunction.TYPE_EQ

        return list(filter(filter_eq, self.constraints))

    def get_ineq_constraints(self):
        """Accessor for all equality constraints

        :returns: a list of equality constraints
        """

        def filter_ineq(cstr):
            """A filter for equality constraints

            :param cstr: constraint function
            :returns: True if the function is an inequality constraint
            """
            return cstr.f_type == MDOFunction.TYPE_INEQ

        return list(filter(filter_ineq, self.constraints))

    def get_observable(self, name):
        """
        Returns the required observable.

        :param name: name of the observable
        :type name: str
        """
        try:
            observable = next(obs for obs in self.observables if obs.name == name)
        except StopIteration:
            raise ValueError("Observable {} cannot be found.".format(name))

        return observable

    def get_ineq_constraints_number(self):
        """Computes the number of inequality constraints

        :returns: the number of inequality constraints
        """
        return len(self.get_ineq_constraints())

    def get_eq_constraints_number(self):
        """Computes the number of equality constraints

        :returns: the number of equality constraints
        """
        return len(self.get_eq_constraints())

    def get_constraints_number(self):
        """Computes the number of equality constraints

        :returns: the number of equality constraints

        """
        return len(self.constraints)

    def get_constraints_names(self):
        """
        Get all constraints names as a list

        :returns: the list of constraints names
        """
        names = [constraint.name for constraint in self.constraints]
        return names

    def get_nonproc_constraints(self):
        """Returns the list of nonprocessed constraints."""
        return self.nonproc_constraints

    def get_design_variable_names(self):
        """Returns a list of all design variables names"""
        return self.design_space.variables_names

    def get_all_functions(self):
        """Returns a list of all functions of the MDO problem
        optimization constraints and objective
        """
        return [self.objective] + self.constraints + self.observables

    def get_all_functions_names(self):
        """
        Get all constraints and objective names

        :returns: a list of names of all functions of the MDO problem
            optimization constraints and objective
        """
        return [func.name for func in self.get_all_functions()]

    def get_objective_name(self):
        """
        Get objective function name

        :returns: the name of the actual objective function
        """
        return self.objective.name

    def get_nonproc_objective(self):
        """Returns the nonprocessed objective."""
        return self.nonproc_objective

    def has_nonlinear_constraints(self):
        """Checks if the problem has constraints

        :returns: True if the problem has equality or inequality constraints
        """
        return len(self.constraints) > 0

    def __notify_store_listeners(self):
        """
        Notifies the listeners that a new store has been made in the
        database
        """
        for func in self.__store_listeners:
            func()

    def __notify_newiter_listeners(self, xvect=None):
        """
        Notifies the listeners that a new iteration is ongoing

        :param xvect: design parameters
        :type xvect: ndarray
        """

        for func in self.__newiter_listeners:
            func()

        if xvect is not None:
            for obs in self.new_iter_observables:
                obs(xvect)

    def has_constraints(self):
        """Checks if the problem has equality or inequality constraints

        :returns: True if the problem has  constraints
        """
        return self.has_eq_constraints() or self.has_ineq_constraints()

    def has_eq_constraints(self):
        """Checks if the problem has equality constraints

        :returns: True if the problem has equality constraints
        """
        return len(self.get_eq_constraints()) > 0

    def has_ineq_constraints(self):
        """Checks if the problem has inequality constraints

        :returns: True if the problem has inequality constraints

        """
        return len(self.get_ineq_constraints()) > 0

    def get_x0_normalized(self):
        """Accessor for the normalized x_0

        :returns: x0-lb)/(ub-lb)
        """
        dspace = self.design_space
        return dspace.normalize_vect(dspace.get_current_x())

    def get_dimension(self):
        """
        Get the total number of design variables

        :returns: the dimension of the design space
        """
        return self.design_space.dimension

    @property
    def dimension(self):
        """dimension property, ie dimension of the design space"""
        return self.design_space.dimension

    @staticmethod
    def check_format(input_function):
        """Checks that the input_function is an istance of MDOFunction

        :param input_function: function
        """
        if not isinstance(input_function, MDOFunction):
            raise TypeError(
                "Optimization problem functions must be "
                + "instances of gemseo.core.MDOFunction"
            )

    def get_eq_cstr_total_dim(self):
        """Returns the total number of equality constraints dimensions
        that is the sum of all outputs dimensions of
        all constraints

        :returns: total number of equality constraints
        """
        return self.__count_cstr_total_dim(MDOFunction.TYPE_EQ)

    def get_ineq_cstr_total_dim(self):
        """Returns the total number of inequality constraints dimensions
        that is the sum of all outputs dimensions of
        all constraints

        :returns: total number of inequality constraints dimensions
        """
        return self.__count_cstr_total_dim(MDOFunction.TYPE_INEQ)

    def __count_cstr_total_dim(self, cstr_type):
        """
        Returns the total number of equality or inequality constraints
        dimensions that is the sum of all outputs dimensions of
        all constraints

        :param cstr_type: the type of contraint, TYPE_INEQ or TYPE_EQ
        :returns: total number of constraints dimensions
        """
        n_cstr = 0
        for constraint in self.constraints:
            if not constraint.has_dim():
                raise ValueError(
                    "Constraint dimension not available yet,"
                    + " please call function "
                    + str(constraint)
                    + " once."
                )
            if constraint.f_type == cstr_type:
                n_cstr += constraint.dim
        return n_cstr

    def get_active_ineq_constraints(self, x_vect, tol=1e-6):
        """
        Returns active constraints names and indices

        :param x_vect: vector of x values, not normalized
        """
        self.design_space.check_membership(x_vect)
        normalize = self.preprocess_options.get("normalize", False)
        if normalize:
            x_vect = self.design_space.normalize_vect(x_vect)

        act_funcs = {}

        for func in self.get_ineq_constraints():
            val = np_abs(func(x_vect))
            act_funcs[func] = where(val <= tol, True, False)

        return act_funcs

    def __wrap_in_database(self, orig_func, normalized=True, is_observable=False):
        """
        Wraps the function to test if it is already in the database
        and store its evaluation

        :param  orig_func: the MDOFunction to be wrapped
        :param  normalized: if True, the input of orig_func are assumed
            normalized; otherwise they are assumed non-normalized
        :param is_observable: if True, new_iter_listeners are not called when
            function is called (avoid recursive call)
        :returns: the wrapped function as an MDOFunction
        """
        fname = orig_func.name
        normalize_vect = self.design_space.normalize_vect
        unnormalize_vect = self.design_space.unnormalize_vect

        def unnormalize_gradient(x_vect):
            """Unnormalize gradient

            :param x_vect: gradient
            :return: unnormalized gradient
            """
            return normalize_vect(x_vect, minus_lb=False)

        def normalize_gradient(x_vect):
            """Normalize gradient

            :param x_vect: gradient
            :return: normalized gradient
            """
            return unnormalize_vect(x_vect, minus_lb=False, no_check=True)

        def wrapped_function(x_vect):
            """Wrapped provided function in order to give to
            optimizer

            :param x_vect: design variable
            :returns: evaluation of function at x_vect
            """
            if np_any(np_isnan(x_vect)):
                raise DesvarIsNan(
                    "Design Variables contain a NaN value !" + str(x_vect)
                )
            if normalized:
                xn_vect = x_vect
                xu_vect = unnormalize_vect(xn_vect)
            else:
                xu_vect = x_vect
                xn_vect = normalize_vect(xu_vect)
            # try to retrieve the evaluation
            value = None
            if not self.database.get(xu_vect, False) and not is_observable:
                if normalized:
                    self.__notify_newiter_listeners(xn_vect)
                else:
                    self.__notify_newiter_listeners(xu_vect)
            else:
                value = self.database.get_f_of_x(fname, xu_vect)
            if value is None:
                # if not evaluated yet, evaluate
                if normalized:
                    value = orig_func(xn_vect)
                else:
                    value = orig_func(xu_vect)
                if self.stop_if_nan and np_any(np_isnan(value)):
                    raise FunctionIsNan(
                        "Function " + str(fname) + " is NaN for x=" + str(xu_vect)
                    )
                values_dict = {fname: value}
                # store (x, f(x)) in database
                self.database.store(xu_vect, values_dict)
                self.__notify_store_listeners()

            return value

        db_func = MDOFunction(
            wrapped_function,
            name=fname,
            f_type=orig_func.f_type,
            expr=orig_func.expr,
            args=orig_func.args,
            dim=orig_func.dim,
            outvars=orig_func.outvars,
        )

        if orig_func.has_jac():

            def dwrapped_function(x_vect):
                """Wrapped provided gradient in order to give to
                optimizer

                :param x_vect: design variable
                :returns: evaluation of gradient at x_vect
                """
                if np_any(np_isnan(x_vect)):
                    raise FunctionIsNan(
                        "Design Variables contain a NaN " + "value !" + str(x_vect)
                    )
                if normalized:
                    xn_vect = x_vect
                    xu_vect = unnormalize_vect(xn_vect)
                else:
                    xu_vect = x_vect
                    xn_vect = normalize_vect(xu_vect)
                # try to retrieve the evaluation
                jac_u = None
                if not self.database.get(xu_vect, False):
                    self.__notify_newiter_listeners()
                else:
                    jac_u = self.database.get_f_of_x(Database.GRAD_TAG + fname, xu_vect)

                if jac_u is not None:
                    jac_n = normalize_gradient(jac_u)
                else:
                    # if not evaluated yet, evaluate
                    if normalized:
                        jac_n = orig_func.jac(xn_vect).real
                        jac_u = unnormalize_gradient(jac_n)
                    else:
                        jac_u = orig_func.jac(xu_vect).real
                        jac_n = normalize_gradient(jac_u)
                    if np_any(np_isnan(jac_n)) and self.stop_if_nan:
                        raise FunctionIsNan(
                            "Function "
                            + str(fname)
                            + " 's Jacobian is NaN for x="
                            + str(xu_vect)
                        )
                    values_dict = {Database.GRAD_TAG + fname: jac_u}
                    # store (x, j(x)) in database
                    self.database.store(xu_vect, values_dict)
                    self.__notify_store_listeners()
                if normalized:
                    jac = jac_n
                else:
                    jac = jac_u
                return jac

            db_func.jac = dwrapped_function

        return db_func

    def add_callback(self, callback_func, each_new_iter=True, each_store=False):
        """Adds a callback function after each store operation or new iteration

        :param callback_func: a function called after the
            function if None nothing
        :param each_new_iter: if True, callback at every iteration
        :param each_store: if True, callback at every call to store()
            in the database
        """
        if each_store:
            self.add_store_listener(callback_func)
        if each_new_iter:
            self.add_new_iter_listener(callback_func)

    def add_store_listener(self, listener_func):
        """
        When an item is stored to the database, calls the listener
        functions

        :param listener_func : function to be called
        :param args: optional arguments of function
        """
        if not callable(listener_func):
            raise TypeError("Listener function is not callable")
        self.__store_listeners.append(listener_func)

    def add_new_iter_listener(self, listener_func):
        """
        When a new iteration stored to the database, calls the listener
        functions

        :param listener_func : function to be called
        :param args: optional arguments of function
        """
        if not callable(listener_func):
            raise TypeError("Listener function is not callable")
        self.__newiter_listeners.append(listener_func)

    def clear_listeners(self):
        """
        Clears all the new_iter and store listeners
        """
        self.__store_listeners = []
        self.__newiter_listeners = []

    def evaluate_functions(
        self,
        x_vect=None,
        eval_jac=False,
        eval_obj=True,
        normalize=True,
        no_db_no_norm=False,
    ):
        """Compute objective and constraints at x_normed
        Some libraries require the number of constraints as an input parameter
        which is unknown by formulation/scenario. Evaluation of initial point
        allows to get this mandatory informations.
        Also used for DOE to evaluate samples

        :param x_normed: the normalized vector at which the
            point must be evaluated if None, x_0 is used (Default value = None)
        :param eval_jac: if True, the jacobian is also evaluated
            (Default value = False)
        :param eval_obj: if True, the objective is evaluated
            (Default value = True)
        :param no_db_no_norm: if True, dont use preprocessed functions,
            so we have no database, nor normalization
        """
        if no_db_no_norm and normalize:
            raise ValueError(
                "Can't use no_db_no_norm " + "and normalize options together"
            )
        if normalize:
            if x_vect is None:
                x_vect = self.get_x0_normalized()
            else:
                # Checks proposed x wrt bounds
                x_u_r = self.design_space.unnormalize_vect(x_vect)
                self.design_space.check_membership(x_u_r)
        else:
            if x_vect is None:
                x_vect = self.design_space.get_current_x()
            else:
                # Checks proposed x wrt bounds
                self.design_space.check_membership(x_vect)

        if no_db_no_norm:
            if eval_obj:
                functions = self.nonproc_constraints + [self.nonproc_objective]
            else:
                functions = self.nonproc_constraints
        else:
            if eval_obj:
                functions = self.constraints + [self.objective]
            else:
                functions = self.constraints

        outputs = {}
        for func in functions:
            try:
                outputs[func.name] = func(x_vect)
            except ValueError:
                LOGGER.error("Failed to evaluate function %s", str(func.name))
                raise

        jacobians = {}
        if eval_jac:
            for func in functions:
                try:
                    jacobians[func.name] = func.jac(x_vect)
                except ValueError:
                    msg = "Failed to evaluate " "jacobian of {}".format(str(func.name))
                    LOGGER.error(msg)
                    raise

        return outputs, jacobians

    def preprocess_functions(self, normalize=True, use_database=True, round_ints=True):
        """Preprocesses all the functions: objective and constraints
        to wrap them with the database and eventually
        the gradients by complex step or FD
        :param normalize: if True, the function is normalized
        :type normalize: bool
        :param use_database: if True, the function is wrapped in the database
        :type use_database: bool
        :param round_ints: if True, rounds integer variables
        :type round_ints: bool
        """
        # Avoids multiple wrappings of functions when multiple executions
        # are performed, in bi level scenarios for instance
        if not self.__functions_are_preprocessed:
            self.preprocess_options = {
                "normalize": normalize,
                "use_database": use_database,
                "round_ints": round_ints,
            }
            # Preprocess the constraints
            for icstr, cstr in enumerate(self.constraints):
                non_p_cstr = self.__preprocess_func(
                    cstr, normalize=False, use_database=False, round_ints=round_ints
                )
                self.nonproc_constraints.append(non_p_cstr)
                p_cstr = self.__preprocess_func(
                    cstr, normalize=normalize, use_database=use_database
                )
                self.constraints[icstr] = p_cstr
            # Preprocess the observables
            for iobs, obs in enumerate(self.observables):
                non_p_obs = self.__preprocess_func(
                    obs,
                    normalize=False,
                    use_database=False,
                    round_ints=round_ints,
                    is_observable=True,
                )
                self.nonproc_observables.append(non_p_obs)
                p_obs = self.__preprocess_func(
                    obs,
                    normalize=normalize,
                    use_database=use_database,
                    is_observable=True,
                )
                self.observables[iobs] = p_obs
            for iobs, obs in enumerate(self.new_iter_observables):
                non_p_obs = self.__preprocess_func(
                    obs,
                    normalize=False,
                    use_database=False,
                    round_ints=round_ints,
                    is_observable=True,
                )
                self.nonproc_new_iter_observables.append(non_p_obs)
                p_obs = self.__preprocess_func(
                    obs,
                    normalize=normalize,
                    use_database=use_database,
                    is_observable=True,
                )
                self.new_iter_observables[iobs] = p_obs
            # Preprocess the objective
            np_o = self.__preprocess_func(
                self.objective,
                normalize=False,
                use_database=False,
                round_ints=round_ints,
            )
            self.nonproc_objective = np_o
            self.objective = self.__preprocess_func(
                self.objective,
                normalize=normalize,
                use_database=use_database,
                round_ints=round_ints,
            )
            self.objective.f_type = MDOFunction.TYPE_OBJ
            self.__functions_are_preprocessed = True
            self.check()

    def __preprocess_func(
        self,
        function,
        normalize=True,
        use_database=True,
        round_ints=True,
        is_observable=False,
    ):
        """
        Wraps the function to: differentiate it and store its call in
        the database. Only the computed gradients are stored in the database,
        not the eventual finite differences or complex step
        perturbed evaluations

        :param function: the scaled and derived MDOFunction
        :param normalize: if True, the function is normalized
        :param use_database: if True, the function is wrapped in the database
        :param round_ints: if True, rounds integer variables before call
        :param is_observable: if True, new_iter_listeners are not called when
            function is called (avoid recursive call)
        :returns: the preprocessed function
        """
        self.check_format(function)
        # First differentiate it so that the finite differences evaluations
        # are not stored in the database, which would be the case in the other
        # way round
        # Also, store non normalized values in the database for further
        # exploitation
        function = self.__normalize_and_round(function, normalize, round_ints)
        self.__add_fd_jac(function)

        # Cast to real value, the results can be a complex number (ComplexStep)
        function.force_real = True
        if use_database:
            function = self.__wrap_in_database(function, normalize, is_observable)
        return function

    def __normalize_and_round(self, orig_func, normalize, round_ints):
        """
        Create a function that takes a scaled input vector
        instead of the original input vector

        :param orig_func: the function as an MDOFunction
        :param normalize: if True, the function is normalized
        :param round_ints: if True, rounds integer variables before call

        :returns: the normalized and differentiated function
        """
        if (not normalize) and (not round_ints):
            return orig_func

        unnormalize_vect = self.design_space.unnormalize_vect

        def normalize_gradient(x_vect):
            """Normalize gradient

            :param x_vect: gradient
            :return: normalized gradient
            """
            return unnormalize_vect(x_vect, minus_lb=False, no_check=True)

        round_int_vars = self.design_space.round_vect

        def f_wrapped(x_vect):
            """Unnormalize design vector for function evaluation

            :param x_vect: normalized design vector
            :returns: function value at x_vect
            """
            if normalize:
                x_vect = unnormalize_vect(x_vect)
            if round_ints:
                x_vect = round_int_vars(x_vect)
            return orig_func(x_vect)

        normed_func = MDOFunction(
            f_wrapped,
            name=orig_func.name,
            f_type=orig_func.f_type,
            expr=orig_func.expr,
            args=orig_func.args,
            dim=orig_func.dim,
            outvars=orig_func.outvars,
        )

        def df_wrapped(x_vect):
            """Unnormalize design vector for gradient evaluation

            :param x_vect: normalized design vector
            :returns: gradient value at xn
            """
            if not orig_func.has_jac():
                raise ValueError(
                    "Selected user gradient "
                    + " but function "
                    + str(orig_func)
                    + " has no Jacobian matrix !"
                )
            if normalize:
                x_vect = unnormalize_vect(x_vect)
            if round_ints:
                x_vect = round_int_vars(x_vect)
            g_u = orig_func.jac(x_vect)
            if normalize:
                return normalize_gradient(g_u)
            return g_u

        normed_func.jac = df_wrapped

        return normed_func

    def __add_fd_jac(self, function):
        """
        Adds a pointer to the jacobian of the function
        generated either by COMPLEX_STEP or FINITE_DIFFERENCES

        :param function: the function to be derivated
        """
        if self.differentiation_method == self.COMPLEX_STEP:
            c_s = ComplexStep(function.evaluate, self.fd_step)
            function.jac = c_s.f_gradient

        if self.differentiation_method == self.FINITE_DIFFERENCES:
            f_d = FirstOrderFD(function, self.fd_step)
            function.jac = f_d.f_gradient

    def check(self):
        """Checks if the optimization problem is ready for run"""
        if self.objective is None:
            raise ValueError("Missing objective function" + " in OptimizationProblem")
        self.__check_pb_type()
        self.design_space.check()
        self.__check_differentiation_method()
        self.check_format(self.objective)
        self.__check_functions()

    def __check_functions(self):
        """
        Checks that the constraints are well declared
        """
        for cstr in self.constraints:
            self.check_format(cstr)
            if not cstr.is_constraint():
                raise ValueError(
                    "Constraint type is not eq or ineq !, got "
                    + str(cstr.f_type)
                    + " instead "
                )
        self.check_format(self.objective)

    def __check_pb_type(self):
        """
        Checks that the pb_type is among self.AVAILABLE_PB_TYPES
        """
        if self.pb_type not in self.AVAILABLE_PB_TYPES:
            raise TypeError(
                "Unknown problem type "
                + str(self.pb_type)
                + ", available problem types are "
                + str(self.AVAILABLE_PB_TYPES)
            )

    def __check_differentiation_method(self):
        """
        Checks that the differentiation method is in allowed ones
        """
        if self.differentiation_method not in self.DIFFERENTIATION_METHODS:
            raise ValueError(
                "Differentiation method "
                + str(self.differentiation_method)
                + " is not among the supported ones : "
                + str(self.DIFFERENTIATION_METHODS)
            )

        if self.differentiation_method == self.COMPLEX_STEP:
            if self.fd_step == 0:
                raise ValueError("ComplexStep step is null !")
            if self.fd_step.imag != 0:
                LOGGER.warning(
                    "Complex step method has an imaginary "
                    "step while required a pure real one."
                    " Auto setting the real part"
                )
                self.fd_step = self.fd_step.imag
        elif self.differentiation_method == self.FINITE_DIFFERENCES:
            if self.fd_step == 0:
                raise ValueError("Finite differences step is null !")
            if self.fd_step.imag != 0:
                LOGGER.warning(
                    "Finite differences method has a complex "
                    "step while required a pure real one."
                    " Auto setting the imaginary part to 0"
                )
                self.fd_step = self.fd_step.real

    def change_objective_sign(self):
        """Changes the objective function sign, when it needs to be maximized
        for instance
        """
        self.minimize_objective = not self.minimize_objective
        self.objective = -self.objective
        # Use MDOFunction Operator overloading

    def _satisfied_constraint(self, cstr_type, value):
        """Determine if an evaluation satisfies a constraint within
        a given tolerance

        :param cstr_type: type of the constraint
        :param value: evaluation of the constraint
        """
        if cstr_type == MDOFunction.TYPE_EQ:
            return np_all(np_abs(value) <= self.eq_tolerance)
        return np_all(value <= self.ineq_tolerance)

    def is_point_feasible(self, out_val, constraints=None):
        """
        Returns True if the point is feasible

        :param out_val: dict of values, containing objective function
            and eventually constraints.
            Warning: if the constraint value is not present,
            the constraint will be considered satisfied
        :param constraints: the list of constraints (MDOFunctions)
            to check. If None, takes all constraints of the problem
        """
        if constraints is None:
            constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        feasible = True
        for constraint in constraints:
            # look for the evaluation of the constraint
            eval_cstr = out_val.get(constraint.name, None)
            # if evaluation exists, check if it is satisfied
            if eval_cstr is None or not self._satisfied_constraint(
                constraint.f_type, eval_cstr
            ):
                feasible = False
                break
        return feasible

    def get_feasible_points(self):
        """Return the list of feasible points within a given tolerance
        eq_tolerance and  ineq_tolerance are taken from sel attrs
        """
        x_history = []
        f_history = []
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()

        for x_vect, out_val in self.database.items():
            feasible = self.is_point_feasible(out_val, constraints=constraints)

            # if all constraints are satisfied, store the vector
            if feasible:
                x_history.append(x_vect.unwrap())
                f_history.append(out_val)
        return x_history, f_history

    def get_violation_criteria(self, x_vect):
        """
        Computes a violation measure associated to an iteration
        For each constraints, when it is violated, add the absolute
        distance to zero, in L2 norm

        if 0, all constraints are satisfied

        :param x_vect: vector of design variables
        :returns: True if feasible, and the violation criteria
        """
        f_violation = 0.0
        is_pt_feasible = True
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        out_val = self.database.get(x_vect)
        for constraint in constraints:
            # look for the evaluation of the constraint
            eval_cstr = out_val.get(constraint.name, None)
            # if evaluation exists, check if it is satisfied
            if eval_cstr is None:
                break
            if not self._satisfied_constraint(constraint.f_type, eval_cstr):
                is_pt_feasible = False
                if constraint.f_type == MDOFunction.TYPE_INEQ:
                    if isinstance(eval_cstr, ndarray):
                        viol_inds = where(eval_cstr > self.ineq_tolerance)
                        f_violation += (
                            norm(eval_cstr[viol_inds] - self.ineq_tolerance) ** 2
                        )
                    else:
                        f_violation += (eval_cstr - self.ineq_tolerance) ** 2
                else:
                    f_violation += norm(abs(eval_cstr) - self.eq_tolerance) ** 2
        return is_pt_feasible, f_violation

    def get_best_infeasible_point(self):
        """Return the best infeasible point within a given tolerance """
        x_history = []
        f_history = []
        is_feasible = []
        viol_criteria = []
        for x_vect, out_val in self.database.items():
            is_pt_feasible, f_violation = self.get_violation_criteria(x_vect)
            is_feasible.append(is_pt_feasible)
            viol_criteria.append(f_violation)
            x_history.append(x_vect.unwrap())
            f_history.append(out_val)

        is_opt_feasible = False
        if viol_criteria:
            best_i = int(argmin(array(viol_criteria)))
            is_opt_feasible = is_feasible[best_i]
        else:
            best_i = 0

        opt_f_dict = {}
        if len(f_history) <= best_i:
            f_opt = None
            x_opt = None
        else:
            f_opt = f_history[best_i].get(self.objective.name)
            x_opt = x_history[best_i]
            opt_f_dict = f_history[best_i]
        return x_opt, f_opt, is_opt_feasible, opt_f_dict

    def __get_optimum_infeas(self):
        """Return the optimum solution within a given feasibility
        tolerances, when there is no feasible point

        :returns: tuple, best evaluation iteration and solution
        """
        msg = (
            "Optimization found no feasible point ! "
            + " The least infeasible point is selected."
        )
        LOGGER.warning(msg)
        x_opt, f_opt, _, f_history = self.get_best_infeasible_point()
        c_opt = {}
        c_opt_grad = {}
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        for constraint in constraints:
            c_opt[constraint.name] = f_history.get(constraint.name)
            f_key = Database.GRAD_TAG + constraint.name
            c_opt_grad[constraint.name] = f_history.get(f_key)
        return x_opt, f_opt, c_opt, c_opt_grad

    def __get_optimum_feas(self, feas_evals, feas_pts):
        """Return the optimum solution within a given feasibility
        tolerances, when there is a feasible point

        :param fea_evals: feasible evaluation list
        :parm feas_pts: feasible points list
        :returns: tuple, best evaluation iteration and solution
        """
        f_opt, x_opt = inf, None
        c_opt = {}
        c_opt_grad = {}
        obj_name = self.objective.name
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        for (i, out_val) in enumerate(feas_evals):
            obj_eval = out_val.get(obj_name)
            if obj_eval is None or isinstance(obj_eval, Number) or obj_eval.size == 1:
                tmp_objeval = obj_eval
            else:
                tmp_objeval = norm(obj_eval)
            if tmp_objeval is not None and tmp_objeval < f_opt:
                f_opt = tmp_objeval
                x_opt = feas_pts[i]
                for constraint in constraints:
                    c_name = constraint.name
                    c_opt[c_name] = feas_evals[i].get(c_name)
                    c_key = Database.GRAD_TAG + c_name
                    c_opt_grad[constraint.name] = feas_evals[i].get(c_key)
        return x_opt, f_opt, c_opt, c_opt_grad

    def get_optimum(self):
        """Return the optimum solution within a given feasibility
        tolerances

        :returns: tuple, best evaluation iteration and solution
        """
        if not self.database:
            raise ValueError("Optimization history is empty")
        feas_pts, feas_evals = self.get_feasible_points()

        if not feas_pts:
            is_feas = False
            x_opt, f_opt, c_opt, c_opt_d = self.__get_optimum_infeas()
        else:
            is_feas = True
            x_opt, f_opt, c_opt, c_opt_d = self.__get_optimum_feas(feas_evals, feas_pts)
        return f_opt, x_opt, is_feas, c_opt, c_opt_d

    def _get_msg(self, max_ds_size=40):
        """
        Get logging message representing self

        :param max_ds_size: maximum design space dimension
            to print
        """
        msg = "Optimization problem:\n"
        msg += "      Minimize: "
        msg += str(self.objective) + "\n"
        variables_names = self.design_space.variables_names
        msg += "With respect to: \n"
        msg += "    " + str(", ".join(variables_names)) + "\n"

        if self.has_constraints():
            msg += "Subject to constraints: \n"
            for eq_c in self.get_eq_constraints():
                cst_exprs = [_f for _f in str(eq_c).split("\n") if _f]
                for expr_spl in cst_exprs:
                    msg += expr_spl + " = 0\n"

            for ieq_c in self.get_ineq_constraints():
                cst_exprs = [_f for _f in str(ieq_c).split("\n") if _f]
                for expr_spl in cst_exprs:
                    msg += expr_spl + " <= 0\n"
        if self.design_space.dimension <= max_ds_size:
            msg += str(self.design_space)
        return msg

    def __str__(self):
        """
        Gets a string representing the optimization problem
        only the formulation constraints are displayed

        :returns: the opt pb representation
        """

        return self._get_msg()

    def log_me(self, max_ds_size=40):
        """Logs a representation of the optimization problem characteristics
        logs self.__repr__ message

        :param max_ds_size: maximum design space dimension
            to print
        """
        msg = self._get_msg(max_ds_size)
        for line in msg.split("\n"):
            LOGGER.info(line)

    @staticmethod
    def __store_h5data(group, data_array, dataset_name, dtype=None):
        """
        Stores an array in a hdf5 file group

        :param group: the group pointer
        :param data_array: the data to stor
        :param dataset_name: name of the dataset to store the array
        """
        if data_array is None:
            return
        if isinstance(data_array, ndarray):
            data_array = data_array.real
        if isinstance(data_array, string_types):
            data_array = string_array([data_array])
        if isinstance(data_array, list):
            all_str = reduce(
                lambda x, y: x or y,
                (isinstance(data, string_types) for data in data_array),
            )
            if all_str:
                data_array = string_array(data_array)
                dtype = data_array.dtype
        group.create_dataset(dataset_name, data=data_array, dtype=dtype)

    @staticmethod
    def __store_attr_h5data(obj, group):
        """
        Stores an object that has a get_data_dict_repr attribute
        in the hdf5 dataset

        :param obj: the object to store
        :param group: the hdf5 group
        """
        data_dict = obj.get_data_dict_repr()
        for attr_name, attr in data_dict.items():
            dtype = None
            is_arr_n = isinstance(attr, ndarray) and issubdtype(attr.dtype, np_number)
            if isinstance(attr, string_types):
                attr = attr.encode("ascii", "ignore")
            elif isinstance(attr, bytes):
                attr = attr.decode()
            elif hasattr(attr, "__iter__") and not is_arr_n:
                if PY3:
                    attr = [
                        att.encode("ascii", "ignore")
                        if isinstance(att, string_types)
                        else att
                        for att in attr
                    ]
                dtype = h5py.special_dtype(vlen=str)
            OptimizationProblem.__store_h5data(group, attr, attr_name, dtype)

    def export_hdf(self, file_path, append=False):
        """Export optimization problem to hdf file.

        :param file_path: file to store the data
        :param append: if True, data is appended to the file if not empty
            (Default value = False)
        """
        LOGGER.info("Export optimization problem to file: %s", str(file_path))
        if append:
            mode = "a"
        else:
            mode = "w"
        h5file = h5py.File(file_path, mode)

        if not append or self.OPT_DESCR_GROUP not in h5file:
            opt_group = h5file.require_group(self.OPT_DESCR_GROUP)
            for attr_name in self.OPTIM_DESCRIPTION:
                attr = getattr(self, attr_name)
                self.__store_h5data(opt_group, attr, attr_name)

            obj_group = h5file.require_group(self.OBJECTIVE_GROUP)
            self.__store_attr_h5data(self.objective, obj_group)

            if self.constraints:
                cstr_group = h5file.require_group(self.CONSTRAINTS_GROUP)
                for cstr in self.constraints:
                    name = cstr.name
                    c_subgroup = cstr_group.require_group(name)
                    self.__store_attr_h5data(cstr, c_subgroup)

            if hasattr(self.solution, "get_data_dict_repr"):
                sol_group = h5file.require_group(self.SOLUTION_GROUP)
                self.__store_attr_h5data(self.solution, sol_group)

        no_designspace = DesignSpace.DESIGN_SPACE_GROUP not in h5file
        h5file.close()

        self.database.export_hdf(file_path, append=True)
        # Design space shall remain the same in append mode
        if not append or no_designspace:
            self.design_space.export_hdf(file_path, append=True)

    @staticmethod
    def import_hdf(file_path, x_tolerance=0.0):
        """Imports optimization history from hdf file

        :param file_path: file to deserialize
        :returns: the read optimization problem
        """
        LOGGER.info("Import optimization problem from file: %s", str(file_path))
        design_space = DesignSpace(file_path)
        opt_pb = OptimizationProblem(design_space, input_database=file_path)
        h5file = h5py.File(file_path, "r")
        try:
            group = h5file[opt_pb.OPT_DESCR_GROUP]
            for attr_name, attr in group.items():
                val = attr.value
                val = val.decode() if isinstance(val, bytes) else val
                setattr(opt_pb, attr_name, val)

            if opt_pb.SOLUTION_GROUP in h5file:
                group = h5file[opt_pb.SOLUTION_GROUP]
                data_dict = {attr_name: attr.value for attr_name, attr in group.items()}
                data_dict = OptimizationProblem.__cast_bytes_dict(data_dict)
                attr = OptimizationResult.init_from_dict_repr(**data_dict)
                opt_pb.solution = attr

            group = h5file[opt_pb.OBJECTIVE_GROUP]
            data_dict = {attr_name: attr.value for attr_name, attr in group.items()}
            data_dict = OptimizationProblem.__cast_bytes_dict(data_dict)
            attr = MDOFunction.init_from_dict_repr(**data_dict)
            # The generated functions can be called at the x stored in
            # the database
            attr.set_pt_from_database(
                opt_pb.database, design_space, jac=True, x_tolerance=x_tolerance
            )
            opt_pb.objective = attr

            if opt_pb.CONSTRAINTS_GROUP in h5file:
                group = h5file[opt_pb.CONSTRAINTS_GROUP]
                for cstr_name in group.keys():
                    sub_group = group[cstr_name]
                    data_dict = {
                        attr_name: attr.value for attr_name, attr in sub_group.items()
                    }
                    data_dict = OptimizationProblem.__cast_bytes_dict(data_dict)
                    attr = MDOFunction.init_from_dict_repr(**data_dict)
                    opt_pb.constraints.append(attr)
        except KeyError as err:
            h5file.close()
            raise KeyError(
                "Invalid database hdf5 file, missing dataset. " + err.args[0]
            )

        h5file.close()
        return opt_pb

    def export_to_dataset(self, name, by_group=True, categorize=True, opt_naming=True):
        """Export the optimization problem to a :class:`.Dataset`.

        :param str name: dataset name.
        :param bool by_group: if True, store the data by group. Otherwise,
            store them by variables. Default: True
        :param bool categorize: distinguish between the different groups of
            variables. Default: True.
        :parma bool opt_naming: use an optimization naming. Default: True.
        """
        dataset = Dataset(name, by_group)

        # Set the different groups
        in_grp = out_grp = dataset.DEFAULT_GROUP
        cache_output_as_input = True
        if categorize:
            if opt_naming:
                in_grp = dataset.DESIGN_GROUP
                out_grp = dataset.FUNCTION_GROUP
            else:
                in_grp = dataset.INPUT_GROUP
                out_grp = dataset.OUTPUT_GROUP
            cache_output_as_input = False

        # Add database inputs
        inputs_names = self.design_space.variables_names
        sizes = self.design_space.variables_sizes
        inputs = array(self.database.get_x_history())
        data = DataConversion.array_to_dict(inputs, inputs_names, sizes)
        for input_name, value in data.items():
            dataset.add_variable(input_name, value, in_grp)

        # Add database outputs
        all_data_names = self.database.get_all_data_names()
        outputs_names = list(
            set(all_data_names) - set(inputs_names) - set([self.database.ITER_TAG])
        )
        fct = []
        for func_name in outputs_names:
            func = self.database.get_func_history(func_name)
            fct.append(func.reshape((inputs.shape[0], -1)))
            sizes.update({func_name: fct[-1].shape[1]})
        outputs = concatenate(fct, axis=1)
        data = DataConversion.array_to_dict(outputs, outputs_names, sizes)
        for output_name, value in data.items():
            dataset.add_variable(
                output_name, value, out_grp, cache_as_input=cache_output_as_input
            )
        return dataset

    @staticmethod
    def __cast_bytes_dict(orig_dict):
        """
        Casts a dict of bytes to string values, if the values are string arrays
        also casts the output to string

        :param orig_dict : initial dict
        :returns: the casted dict
        """
        casted = {
            key: val.decode() if isinstance(val, bytes) else val
            for key, val in orig_dict.items()
        }

        for key, value in casted.items():
            if isinstance(value, ndarray) and value.dtype.type is np_string:
                if value.size == 1:
                    casted[key] = value[0]
                else:
                    casted[key] = value.tolist()
        return casted



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Distributions based on OpenTURNS
================================

Overview
--------

The :class:`.OTDistribution` class is a concrete class
inheriting from :class:`.Distribution` which is an abstract one.
OT stands for `OpenTURNS <http://www.openturns.org/>`_
which is the library it relies on.

Similarly, the :class:`.OTComposedDistribution` class is a concrete class
inheriting from :class:`.ComposedDistribution` which is an abstract one.

Construction
------------

The :class:`.OTDistribution` of a given uncertain variable is built
from mandatory arguments:

- a variable name,
- a distribution name recognized by OpenTURNS,
- a set of parameters provided as a tuple
  of positional arguments filled in the order
  specified by the OpenTURNS constructor of this distribution.

.. warning::

    The distribution parameters must be provided according to the signature
    of the openTURNS classes. `Access the openTURNS documentation
    <http://openturns.github.io/openturns/latest/user_manual/
    probabilistic_modelling.html>`_.

The constructor has also optional arguments:

- a variable dimension (default: 1),
- a standard representation of these parameters
  (default: use the parameters provided in the tuple),
- a transformation of the variable (default: no transformation),
- lower and upper bounds for truncation (default: no truncation),
- a threshold for the OpenTURNS truncation tool
  (`more details <http://openturns.github.io/openturns/latest/user_manual/
  _generated/openturns.TruncatedDistribution.html>`_).

Classical distributions
-----------------------

This module also implements a deliberately limited selection
of standard probability distributions
in a user-friendly way: :class:`.OTExponentialDistribution`,
:class:`.OTNormalDistribution`, :class:`.OTTriangularDistribution`,
and :class:`.OTUniformDistribution`. More precisely,
the argument whose nature is a tuple of positional parameters
is replaced with several user-defined keyword arguments.
In this way, the use writes :code:`OTUniformDistribution('x', -1., 3.)`
or :code:`OTUniformDistribution('x', lower=-1., upper=3.)`
instead of :code:`OTDistribution('x', 'Normal', (-1., 3.))`.
Furthermore, these classes inheriting from :class:`.OTDistribution`
are documented in such a way that a newbie could easily apprehend them.
"""
from __future__ import absolute_import, division, unicode_literals

import openturns as ots
from future import standard_library
from numpy import array, inf

from gemseo.uncertainty.distributions.distribution import Distribution
from gemseo.uncertainty.distributions.ot_cdist import OTComposedDistribution

standard_library.install_aliases()

OT_WEBSITE = "http://openturns.github.io/openturns/latest/"
OT_WEBSITE += "user_manual/probabilistic_modelling.html"

from gemseo import LOGGER


class OTDistribution(Distribution):
    """ Interface to an OpenTURNS distribution. """

    COMPOSED_DISTRIBUTION = OTComposedDistribution

    def __init__(
        self,
        variable,
        distribution,
        parameters,
        dimension=1,
        standard_parameters=None,
        transformation=None,
        l_b=None,
        u_b=None,
        threshold=0.5,
    ):
        """Constructor

        :param str variable: variable name.
        :param str distribution: distribution name.
        :param tuple parameters: distribution parameters.
        :param int dimension: variable dimension.
        :param dict standard_parameters: standard parameters.
        :param str transformation: standard variable transformation,
            e.g. 'sin(x)'. If None, no transformation.  Default: None.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        """
        super(OTDistribution, self).__init__(
            variable, distribution, parameters, dimension, standard_parameters
        )
        self.marginals = self.__create_distributions(
            self.distribution_name, self.parameters, transformation, l_b, u_b, threshold
        )
        self.distribution = ots.ComposedDistribution(self.marginals)
        math_support = str(self.support)
        LOGGER.info("|_ Mathematical support: %s", math_support)
        num_range = str(self.range)
        LOGGER.info("|_ Numerical range: %s", num_range)
        LOGGER.info("|_ Transformation: %s", self.transformation)

    def get_sample(self, n_samples=1):
        """Get several samples.

        :param int n_samples: number of samples.
        :return: samples
        :rtype: list(array)
        """
        sample = array(self.distribution.getSample(n_samples))
        return sample

    def cdf(self, vector):
        """Evaluate the cumulative density function of the random variable
        marginals for a given instance.

        :param array vector: instance of the random variable.
        :return: cdf values
        :rtype: array
        """
        return array(
            [
                self.marginals[index].computeCDF(ots.Point([value]))
                for index, value in enumerate(vector)
            ]
        )

    def inverse_cdf(self, vector):
        """Evaluate the inverses of the cumulative density functions of the
        random variable marginals for a given unit vector .

        :param array vector: vector of values comprised between 0 and 1 with
            same dimension as the random variable.
        :return: inverse cdf values
        :rtype: array
        """
        return array(
            [
                self.marginals[index].computeQuantile(value)[0]
                for index, value in enumerate(vector)
            ]
        )

    def _pdf(self, index):
        """Get the probability density function of a marginal.

        :param int index: marginal index.
        :return: probability density function
        :rtype: function
        """

        def pdf(point):
            """Probability Density Function (PDF).

            :param float point: point value.
            :return: PDF value.
            :rtype: float
            """
            return self.marginals[index].computePDF(point)

        return pdf

    def _cdf(self, index):
        """Get the cumulative density function of a marginal.

        :param int index: marginal index.
        :return: cumulative density function
        :rtype: function
        """

        def cdf(level):
            """Cumulative Density Function (CDF).

            :param float point: point value.
            :return: CDF value.
            :rtype: float
            """
            return self.marginals[index].computeCDF(level)

        return cdf

    @property
    def mean(self):
        """Get the mean of the random variable.

        :return: mean of the random variable.
        :rtype: array
        """
        return array(self.distribution.getMean())

    @property
    def standard_deviation(self):
        """Get the standard deviation of the random variable.

        :return: standard deviation of the random variable.
        :rtype: array
        """
        return array(self.distribution.getStandardDeviation())

    def __create_distributions(
        self, distribution, parameters, transformation, l_b, u_b, threshold
    ):
        """For each variable dimension, instantiate an openturns distribution
        from its class name and parameters.

        :param str distribution: distribution name,
            ie its openturns class name.
        :param parameters: parameters of the openturns distribution
        :param str transformation: standard variable transformation,
            e.g. 'sin(x)'. If None, no transformation.  Default: None.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        """
        try:
            ot_dist = getattr(ots, distribution)
        except Exception:
            raise ValueError(
                distribution + " is an unknown openturns" " distribution. "
            )
        try:
            ot_dist = [ot_dist(*parameters)] * self.dimension
        except Exception:
            raise ValueError(
                distribution + " does not take these arguments. "
                "More details on: " + OT_WEBSITE
            )
        self.__set_bounds(ot_dist)
        if transformation is not None:
            ot_dist = self.__transform_marginal_dist(ot_dist, transformation)
        self.__set_bounds(ot_dist)
        if l_b is not None or u_b is not None:
            ot_dist = self.__truncate_marginal_dist(ot_dist, l_b, u_b, threshold)
        self.__set_bounds(ot_dist)
        return ot_dist

    def __transform_marginal_dist(self, marginals, transformation):
        """Apply standard transformations on the marginals, e.g. -, +, *,
        **, sin, exp, log, ...

        :param marginals: marginal distributions.
        :param str transformation: mathematical expression, e.g. 'sin(2*x)'
        """
        variable_name = self.variable_name
        transformation = transformation.replace(" ", "")
        symbolic_function = ots.SymbolicFunction([self.variable_name], [transformation])
        marginals = [
            ots.CompositeDistribution(symbolic_function, marginal)
            for marginal in marginals
        ]
        prev = self.transformation
        transf = transformation.replace(variable_name, "(" + prev + ")")
        self.transformation = transf
        return marginals

    def __truncate_marginal_dist(self, distributions, l_b, u_b, threshold=0.5):
        """Truncate the distribution of a random variable.

        :param distributions: openturns distributions.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        :return: transformed openturns distributions.
        """
        marginals = [
            self.__truncate_distribution(dist, index, l_b, u_b, threshold)
            for index, dist in enumerate(distributions)
        ]
        prev = self.transformation
        self.transformation = "Trunc(" + prev + ")"
        return marginals

    def __truncate_distribution(self, distributions, index, l_b, u_b, threshold=0.5):
        """Truncate a distribution with lower bound, upper bound or both.

        :param ot_dist: openturns distributions
        :param float l_b: lower bound value.
        :param float u_b: upper bound value.
        :param float threshold: threshold value in [0,1].
        :return: truncated openturns distributions.
        """
        if l_b is None:
            LOGGER.info(
                "Truncate distribution of component %s" " above %s.", index, u_b
            )
            upper = ots.TruncatedDistribution.UPPER
            current_u_b = self.math_upper_bound[index]
            if u_b > current_u_b:
                raise ValueError("u_b is greater " "than the current upper bound.")
            distributions = ots.TruncatedDistribution(
                distributions, u_b, upper, threshold
            )
        elif u_b is None:
            LOGGER.info(
                "Truncate distribution of component %s" " below %s.", index, l_b
            )
            lower = ots.TruncatedDistribution.LOWER
            current_l_b = self.math_lower_bound[index]
            if l_b < current_l_b:
                raise ValueError("l_b is lower " "than the current lower bound.")
            distributions = ots.TruncatedDistribution(
                distributions, l_b, lower, threshold
            )
        else:
            LOGGER.info(
                "Truncate distribution of component %s" " below %s and above %s.",
                index,
                l_b,
                u_b,
            )
            current_l_b = self.math_lower_bound[index]
            current_u_b = self.math_upper_bound[index]
            if l_b < current_l_b:
                raise ValueError("l_b is lower " "than the current lower bound.")
            if u_b > current_u_b:
                raise ValueError("u_b is greater " "than the current upper bound.")
            distributions = ots.TruncatedDistribution(
                distributions, l_b, u_b, threshold
            )
        return distributions

    def __set_bounds(self, distributions):
        """Set mathematical and numerical bounds (= support and range).

        :param distributions: list of openturns distributions.
        """
        self.math_lower_bound = []
        self.math_upper_bound = []
        self.num_lower_bound = []
        self.num_upper_bound = []
        for distribution in distributions:
            dist_range = distribution.getRange()
            l_b = dist_range.getLowerBound()[0]
            u_b = dist_range.getUpperBound()[0]
            self.num_lower_bound.append(l_b)
            self.num_upper_bound.append(u_b)
            if not dist_range.getFiniteLowerBound()[0]:
                l_b = -inf
            if not dist_range.getFiniteUpperBound()[0]:
                u_b = inf
            self.math_lower_bound.append(l_b)
            self.math_upper_bound.append(u_b)
        self.math_lower_bound = array(self.math_lower_bound)
        self.math_upper_bound = array(self.math_upper_bound)
        self.num_lower_bound = array(self.num_lower_bound)
        self.num_upper_bound = array(self.num_upper_bound)


class OTNormalDistribution(OTDistribution):
    """ Create a normal distribution. """

    def __init__(
        self,
        variable,
        mu=0.0,
        sigma=1.0,
        dimension=1,
        transformation=None,
        l_b=None,
        u_b=None,
        threshold=0.5,
    ):
        """Constructor.

        :param str variable: variable name.
        :param float mu: mean.
        :param float sigma: standard deviation.
        :param int dimension: dimension.
        :param str transformation: standard variable transformation,
            e.g. 'sin(x)'. If None, no transformation.  Default: None.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        """
        standard_parameters = {self.MU: mu, self.SIGMA: sigma}
        super(OTNormalDistribution, self).__init__(
            variable,
            "Normal",
            (mu, sigma),
            dimension,
            standard_parameters,
            transformation,
            l_b,
            u_b,
            threshold,
        )


class OTUniformDistribution(OTDistribution):
    """ Create a uniform distribution. """

    def __init__(
        self,
        variable,
        lower=0.0,
        upper=1.0,
        dimension=1,
        transformation=None,
        l_b=None,
        u_b=None,
        threshold=0.5,
    ):
        """Constructor.

        :param str variable: variable name.
        :param float lower: lower bound.
        :param float upper: upper bound.
        :param int dimension: dimension.
        :param str transformation: standard variable transformation,
            e.g. 'sin(x)'. If None, no transformation.  Default: None.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        """
        standard_parameters = {self.LOWER: lower, self.UPPER: upper}
        super(OTUniformDistribution, self).__init__(
            variable,
            "Uniform",
            (lower, upper),
            dimension,
            standard_parameters,
            transformation,
            l_b,
            u_b,
            threshold,
        )


class OTTriangularDistribution(OTDistribution):
    """ Create a triangular distribution. """

    def __init__(
        self,
        variable,
        lower=0.0,
        mode=0.5,
        upper=1.0,
        dimension=1,
        transformation=None,
        l_b=None,
        u_b=None,
        threshold=0.5,
    ):
        """Constructor.

        :param str variable: variable name.
        :param float lower: lower bound.
        :param float mode: mode.
        :param float upper: upper bound.
        :param int dimension: dimension.
        :param str transformation: standard variable transformation,
            e.g. 'sin(x)'. If None, no transformation.  Default: None.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        """
        standard_parameters = {self.LOWER: lower, self.MODE: mode, self.UPPER: upper}
        super(OTTriangularDistribution, self).__init__(
            variable,
            "Triangular",
            (lower, mode, upper),
            dimension,
            standard_parameters,
            transformation,
            l_b,
            u_b,
            threshold,
        )


class OTExponentialDistribution(OTDistribution):
    """ Create a exponential distribution. """

    def __init__(
        self,
        variable,
        rate=1.0,
        loc=0.0,
        dimension=1,
        transformation=None,
        l_b=None,
        u_b=None,
        threshold=0.5,
    ):
        """Constructor.

        :param str variable: variable name.
        :param float rate: rate parameter.
        :param float loc: location parameter.
        :param int dimension: dimension.
        :param str transformation: standard variable transformation,
            e.g. 'sin(x)'. If None, no transformation.  Default: None.
        :param float l_b: lower bound for truncation. If None, no lower
            truncation. Default: None.
        :param float u_b: upper bound for truncation. If None, no upper
            truncation. Default: None.
        :param float threshold: threshold value in [0,1].
        """
        standard_parameters = {self.RATE: rate, self.LOC: loc}
        super(OTExponentialDistribution, self).__init__(
            variable,
            "Exponential",
            (rate, loc),
            dimension,
            standard_parameters,
            transformation,
            l_b,
            u_b,
            threshold,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#         documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Benchmark MDO formulations based on scalable disciplines
========================================================

The :mod:`~gemseo.problems.scalable.api` implements several classes
to benchmark MDO formulations based on scalable disciplines.

The :class:`.ScalabilityStudy` class implements
the concept of scalability study:

1. By instantiating a :class:`.ScalabilityStudy`, the user defines
   the MDO problem in terms of design parameters, objective function and
   constraints.
2. For each discipline, the user adds a dataset stored
   in a :class:`.AbstractFullCache` and select a type of
   :class:`.ScalableModel` to build the :class:`.ScalableDiscipline`
   associated with this discipline.
3. The user adds different optimization strategies, defined in terms
   of both optimization algorithms and MDO formulation.
4. The user adds different scaling strategies, in terms of sizes of
   design parameters, coupling variables and equality and inequality
   constraints. The user can also define a scaling strategies according to
   particular parameters rather than groups of parameters.
5. Lastly, the user executes the :class:`.ScalabilityStudy` and the results
   are written in several files and stored into directories
   in a hierarchical way, where names depends on both MDO formulation,
   scaling strategy and replications when it is necessary. Different kinds
   of files are stored: optimization graphs, dependency matrix plots and
   of course, scalability results by means of a dedicated class:
   :class:`.ScalabilityResult`.

The :class:`.PostScalabilityStudy` class implements the way as the set of
:class:`.ScalabilityResult`-based result files
contained in the study directory are graphically post-processed. This class
provides several methods to easily change graphical properties, notably
the plot labels. It also makes it possible to define a cost function per
MDO formulation, converting the numbers of executions and linearizations
of the different disciplines required by a MDO process in an estimation
of the computational cost associated with what would be a scaled version
of the true problem.

.. warning::

   Comparing MDO formulations in terms of estimated true computational time
   rather than CPU time of the :class:`.ScalabilityStudy` is highly
   recommended.
   Indeed, time is often an obviousness criterion to distinguish between
   MDO formulations having the same performance in terms of distance to the
   optimum: look at our calculation budget and choose the best formulation
   that satisfies this budget, or even saves us time. Thus, it is important
   to carefully define these cost functions.

"""
from __future__ import absolute_import, division, unicode_literals

import numbers
import os
import pickle
from builtins import int, open
from copy import deepcopy

import matplotlib.pyplot as plt
from future import standard_library
from matplotlib.lines import Line2D
from numpy import array, inf, median
from past.builtins import basestring

from gemseo.problems.scalable.problem import ScalableProblem

standard_library.install_aliases()

from gemseo import LOGGER

CURRENT_DIRECTORY = os.getcwd()
RESULTS_DIRECTORY = "results"
POST_DIRECTORY = "visualization"
POSTOPTIM_DIRECTORY = os.path.join(POST_DIRECTORY, "optimization_history")
POSTSTUDY_DIRECTORY = os.path.join(POST_DIRECTORY, "scalability_study")
POSTSCAL_DIRECTORY = os.path.join(POST_DIRECTORY, "dependency_matrix")


class ScalabilityStudy(object):
    """Scalability Study."""

    def __init__(
        self,
        objective,
        design_variables,
        directory="study",
        prefix="",
        eq_constraints=None,
        ineq_constraints=None,
        maximize_objective=False,
        fill_factor=0.7,
        active_probability=0.1,
        feasibility_level=0.8,
        start_at_equilibrium=True,
        early_stopping=True,
        coupling_variables=None,
    ):
        """
        The constructor of the ScalabilityStudy class requires two mandatory
        arguments:

        - the :code:`'objective'` name,
        - the list of :code:`'design_variables'` names.

        Concerning output files, we can specify:

        - the :code:`directory` which is :code:`'study'` by default,
        - the prefix of output file names (default: no prefix).

        Regarding optimization parametrization, we can specify:

        - the list of equality constraints names (:code:`eq_constraints`),
        - the list of inequality constraints names (:code:`ineq_constraints`),
        - the choice of maximizing the objective function
          (:code:`maximize_objective`).

        By default, the objective function is minimized and the MDO problem
        is unconstrained.

        Last but not least, with regard to the scalability methodology,
        we can overwrite:

        - the default fill factor of the input-output dependency matrix
          :code:`ineq_constraints`,
        - the probability to set the inequality constraints as active at
          initial step of the optimization :code:`active_probability`,
        - the offset of satisfaction for inequality constraints
          :code:`feasibility_level`,
        - the use of a preliminary MDA to start at equilibrium
          :code:`start_at_equilibrium`,
        - the post-processing of the optimization database to get results
          earlier than final step :code:`early_stopping`.

        :param str objective: name of the objective
        :param list(str) design_variables: names of the design variables
        :param str directory: working directory of the study. Default: 'study'.
        :param str prefix: prefix for the output filenames. Default: ''.
        :param list(str) eq_constraints: names of the equality constraints.
            Default: None.
        :param list(str) ineq_constraints: names of the inequality constraints
            Default: None.
        :param bool maximize_objective: maximizing objective. Default: False.
        :param float fill_factor: default fill factor of the input-output
            dependency matrix. Default: 0.7.
        :param float active_probability: probability to set the inequality
            constraints as active at initial step of the optimization.
            Default: 0.1
        :param float feasibility_level: offset of satisfaction for inequality
            constraints. Default: 0.8.
        :param bool start_at_equilibrium: start at equilibrium
            using a preliminary MDA. Default: True.
        :param bool early_stopping: post-process the optimization database
            to get results earlier than final step.
        """
        LOGGER.info("*** Start a scalability study ***")
        self.prefix = prefix
        self.directory = directory
        self.__create_directories()
        self.datasets = []
        self.objective = objective
        self.design_variables = design_variables
        if eq_constraints is None:
            eq_constraints = []
        self.eq_constraints = eq_constraints
        if ineq_constraints is None:
            ineq_constraints = []
        self.ineq_constraints = ineq_constraints
        if coupling_variables is None:
            coupling_variables = []
        self.coupling_variables = coupling_variables
        self.maximize_objective = maximize_objective
        self.formulations = []
        self.formulations_options = []
        self.algorithms = []
        self.algorithms_options = []
        self.scalings = []
        self.var_scalings = []
        self.__check_fill_factor(fill_factor)
        self._default_fill_factor = fill_factor
        self.__check_proportion(active_probability)
        self.active_probability = active_probability
        if not hasattr(feasibility_level, "__len__"):
            self.__check_proportion(feasibility_level)
        else:
            for _, value in feasibility_level.items():
                self.__check_proportion(value)
        self.feasibility_level = feasibility_level
        self.start_at_equilibrium = start_at_equilibrium
        self.results = []
        self.early_stopping = early_stopping
        self._group_dep = {}
        self._fill_factor = {}
        self.top_level_diff = []
        self._all_data = None
        LOGGER.info("Initialize the scalability study:")
        optimize = "maximize" if self.maximize_objective else "minimize"
        LOGGER.info("|    Objective: %s %s", optimize, self.objective)
        tmp = ", ".join(self.design_variables)
        LOGGER.info("|    Design variables: %s", tmp)
        tmp = ", ".join(self.eq_constraints)
        LOGGER.info("|    Equality constraints: %s", tmp)
        tmp = ", ".join(self.ineq_constraints)
        LOGGER.info("|    Inequality constraints: %s", tmp)
        LOGGER.info("|    Default fill factor: %s", self._default_fill_factor)
        LOGGER.info("|    Active probability: %s", self.active_probability)
        LOGGER.info("|    Feasibility level: %s", self.feasibility_level)
        LOGGER.info("|    Start at equilibrium: %s", self.start_at_equilibrium)
        LOGGER.info("|    Early stopping: %s", self.early_stopping)

    def __create_directories(self):
        """Create the different directories
        to store results, post-processings, ..."""
        LOGGER.info("Create directories:")
        self.__create_directory(self.directory)
        LOGGER.info("|    Working directory: %s", self.directory)
        directory = os.path.join(self.directory, POST_DIRECTORY)
        self.__create_directory(directory)
        LOGGER.info("|    Post-processing: %s", directory)
        directory = os.path.join(self.directory, POSTOPTIM_DIRECTORY)
        self.__create_directory(directory)
        LOGGER.info("|    Optimization history view: %s", directory)
        directory = os.path.join(self.directory, POSTSTUDY_DIRECTORY)
        self.__create_directory(directory)
        LOGGER.info("|    Scalability views: %s", directory)
        directory = os.path.join(self.directory, POSTSCAL_DIRECTORY)
        self.__create_directory(directory)
        LOGGER.info("|    Dependency matrices: %s", directory)
        directory = os.path.join(self.directory, RESULTS_DIRECTORY)
        self.__create_directory(directory)
        LOGGER.info("|    Results: %s", directory)

    @staticmethod
    def __create_directory(path):
        """Create a directory from its relative or absolute path.

        :param str path: relative or absolute directory path.
        """
        if not os.path.exists(path):
            os.mkdir(path)

    def add_discipline(self, name, data):
        """This method adds a disciplinary dataset from:

        - its :code:`name`,
        - its :code:`data`,

        :param str name: name of the discipline.
        :param AbstractFullCache data: dataset provided as a cache.
        """
        self._group_dep[name] = {}
        self._all_data = data.get_all_data()
        datum = self._all_data[1]
        outputs_names = [str(fct_name) for fct_name in list(datum["outputs"].keys())]
        outputs_sizes = [
            str(fct_name) + "(" + str(len(fct_val)) + ")"
            for fct_name, fct_val in datum["outputs"].items()
        ]
        inputs_sizes = [
            str(fct_name) + "(" + str(len(fct_val)) + ")"
            for fct_name, fct_val in datum["inputs"].items()
        ]
        self.datasets.append(data)
        for output_name in outputs_names:
            self.set_fill_factor(name, output_name, self._default_fill_factor)
        LOGGER.info("Add a discipline:")
        LOGGER.info("|    Name: %s", name)
        LOGGER.info("|    Inputs: %s", ", ".join(inputs_sizes))
        LOGGER.info("|    Outputs: %s", ", ".join(outputs_sizes))
        LOGGER.info("|    Sample size: %s", data.get_length())
        LOGGER.info("|    Cache: %s", data.name)

    @property
    def disciplines_names(self):
        """Get discipline names.

        :return: list of discipline names
        :rtype: list(str)
        """
        disc_names = [discipline.name for discipline in self.datasets]
        return disc_names

    def set_input_output_dependency(self, discipline, output, inputs):
        """Set the dependency between an output and a set of inputs
        for a given discipline.

        :param str discipline: name of the discipline
        :param str output: name of the output
        :param list(str) inputs: list of inputs names
        """
        self.__check_discipline(discipline)
        self.__check_output(discipline, output)
        self.__check_inputs(discipline, inputs)
        self._group_dep[discipline][output] = inputs

    def set_fill_factor(self, discipline, output, fill_factor):
        """
        :param str discipline: name of the discipline
        :param str output: name of the output function
        :param float fill_factor: fill factor
        """
        self.__check_discipline(discipline)
        self.__check_output(discipline, output)
        self.__check_fill_factor(fill_factor)
        if discipline not in self._fill_factor:
            self._fill_factor[discipline] = {}
        self._fill_factor[discipline][output] = fill_factor

    def __check_discipline(self, discipline):
        """Check if discipline is a string
        comprised in the list of disciplines names.
        """
        if not isinstance(discipline, basestring):
            raise TypeError("The argument discipline should be a string")
        disciplines_names = self.disciplines_names
        if discipline not in disciplines_names:
            raise ValueError(
                "The argument discipline should be a string " "comprised in the list ",
                disciplines_names,
            )

    def __check_output(self, discipline, output):
        """ Check if output is an output function of discipline. """
        self.__check_discipline(discipline)
        if not isinstance(output, basestring):
            raise TypeError("The argument output should be a string")
        outputs_names = next(
            cache.outputs_names for cache in self.datasets if cache.name == discipline
        )
        if output not in outputs_names:
            raise ValueError(
                "The argument output should be a string " "comprised in the list ",
                outputs_names,
            )

    def __check_inputs(self, discipline, inputs):
        """ Check if inputs is a list of inputs of discipline. """
        self.__check_discipline(discipline)
        inputs_names = next(
            cache.inputs_names for cache in self.datasets if cache.name == discipline
        )
        if not isinstance(inputs, list):
            raise TypeError("The argument inputs must be a list of string.")
        for inpt in inputs:
            self.__check_input(inpt, inputs_names)

    @staticmethod
    def __check_input(inpt, inputs_names):
        """ Check if input is a string comprised in inputs_names. """
        if not isinstance(inpt, basestring):
            raise TypeError("The argument inpt must be a string.")
        if inpt not in inputs_names:
            raise ValueError(
                "The argument inpt must be a string " "comprised in the list ",
                inputs_names,
            )

    def __check_fill_factor(self, fill_factor):
        """Check if fill factor is a proportion or a number equal to -1

        :param float fill_factor: a proportion or -1
        """
        try:
            self.__check_proportion(fill_factor)
        except ValueError:
            if fill_factor != -1:
                raise TypeError(
                    "Fill factor should be "
                    "a float number comprised in 0 and 1 "
                    "or a number equal to -1."
                )

    @staticmethod
    def __check_proportion(proportion):
        """Check if a proportion is a float number comprised in 0 and 1.

        :param float proportion: proportion comprised in 0 and 1.
        """
        if not isinstance(proportion, numbers.Number):
            raise TypeError(
                "A proportion should be " "a float number comprised in 0 and 1."
            )
        if proportion < 0 or proportion > 1:
            raise ValueError(
                "A proportion should be " "a float number comprised in 0 and 1."
            )

    def add_optimization_strategy(
        self,
        algo,
        max_iter,
        formulation="DisciplinaryOpt",
        algo_options=None,
        formulation_options=None,
        top_level_diff="auto",
    ):
        """Add both optimization algorithm and MDO formulation,
        as well as their options.

        :param str algo: name of the optimization algorithm.
        :param int max_iter: maximum number of iterations
            for the optimization algorithm.
        :param str formulation: name of the MDO formulation.
            Default: 'DisciplinaryOpt'.
        :param dict algo_options: options of the optimization algorithm.
        :param dict formulation_options: options of the MDO formulation.
        :param str top_level_diff: differentiation method
            for the top level disciplines. Default: 'auto'.
        """
        self.algorithms.append(algo)
        if algo_options is None:
            algo_options = {}
        else:
            if not isinstance(algo_options, dict):
                raise TypeError("algo_options must be a dictionary.")
        algo_options.update({"max_iter": max_iter})
        self.algorithms_options.append(algo_options)
        self.formulations.append(formulation)
        self.formulations_options.append(formulation_options)
        self.top_level_diff.append(top_level_diff)

    def print_optimization_strategies(self):
        """ Print the list of the optimization strategies. """
        index = 0
        print("Optimization strategies:")
        for algo, formulation in zip(self.algorithms, self.formulations):
            template = "# {} - Algorithm: {} {} - Formulation: {} {}"
            form_opt = self.formulations_options[index]
            form_opt = "" if form_opt is None else form_opt
            algo_opt = self.algorithms_options[index]
            algo_opt = "" if algo_opt is None else algo_opt
            index += 1
            print(template.format(index, algo, algo_opt, formulation, form_opt))

    def add_scaling_strategies(
        self,
        design_size=None,
        coupling_size=None,
        eq_cstr_size=None,
        ineq_cstr_size=None,
        variables=None,
    ):
        """Add different scaling strategies.

        :param design_size: size of the design variables. Default: None.
        :type design_size: int or list(int)
        :param coupling_size: size of the coupling variables. Default: None.
        :type coupling_size: int or list(int)
        :param eq_cstr_size: size of the equality constraints. Default: None.
        :type eq_cstr_size: int or list(int)
        :param ineq_cstr_size: size of the inequality constraints.
            Default: None.
        :type ineq_cstr_size: int or list(int)
        """
        n_design = self.__check_varsizes_type(design_size)
        n_coupling = self.__check_varsizes_type(coupling_size)
        n_eq = self.__check_varsizes_type(eq_cstr_size)
        n_ineq = self.__check_varsizes_type(ineq_cstr_size)
        n_var = self.__check_varsizes_type(variables)
        n_scaling = max(n_design, n_coupling, n_eq, n_ineq, n_var)
        self.__check_scaling_consistency(n_design, n_scaling)
        self.__check_scaling_consistency(n_coupling, n_scaling)
        self.__check_scaling_consistency(n_eq, n_scaling)
        self.__check_scaling_consistency(n_ineq, n_scaling)
        design_size = self.__format_scaling(design_size, n_scaling)
        coupling_size = self.__format_scaling(coupling_size, n_scaling)
        eq_cstr_size = self.__format_scaling(eq_cstr_size, n_scaling)
        ineq_cstr_size = self.__format_scaling(ineq_cstr_size, n_scaling)

        for idx in range(n_scaling):
            var_scaling = {}
            self.__update_var_scaling(
                var_scaling, design_size[idx], self.design_variables
            )
            self.__update_var_scaling(
                var_scaling, coupling_size[idx], self.coupling_variables
            )
            self.__update_var_scaling(
                var_scaling, eq_cstr_size[idx], self.eq_constraints
            )
            self.__update_var_scaling(
                var_scaling, ineq_cstr_size[idx], self.ineq_constraints
            )
            scaling = {
                "design_variables": design_size[idx],
                "coupling_variables": coupling_size[idx],
                "eq_constraint_size": eq_cstr_size[idx],
                "ineq_constraint_size": ineq_cstr_size[idx],
            }
            if variables is not None:
                for varname, value in variables[idx].items():
                    self.__update_var_scaling(var_scaling, value, [varname])
                    scaling[varname] = value
            self.var_scalings.append(var_scaling)
            self.scalings.append(scaling)

    @staticmethod
    def __format_scaling(size, n_scaling):
        """Convert a scaling size in a list of integers
        whose length is equal to the number of scalings.

        :param size: size(s) of a given variable
        :type size: int or list(int)
        :param int n_scaling: number of scalings
        :return: formatted sizes
        """
        formatted_sizes = size
        if isinstance(formatted_sizes, int) or formatted_sizes is None:
            formatted_sizes = [formatted_sizes]
        if len(formatted_sizes) == 1:
            formatted_sizes = formatted_sizes * n_scaling
        return formatted_sizes

    @staticmethod
    def __update_var_scaling(scaling, size, varnames):
        """Update a scaling dictionary for a given list of variables
        and a given size.

        :param dict scaling: scaling dictionary whose keys are variable names
            and values are dictionary with scaling properties,
            e.g. {'size': val}
        :param int size: size of the variable
        :param list(str) varnames: list of variable names
        """
        if size is not None:
            scaling.update({varname: size for varname in varnames})

    @staticmethod
    def __check_scaling_consistency(n_var_scaling, n_scaling):
        """Check that for the different types of variables,
        the number of scalings is the same or equal to 1.

        :param int n_var_scaling: number of scalings
        :param int n_scaling: expected number of scalings
        """
        assert n_var_scaling in (n_scaling, 1)

    @staticmethod
    def __check_varsizes_type(varsizes):
        """Check the type of scaling sizes. Integer, list of integers
        or None is expected. Return the number of scalings.

        :return: length of scalings
        """
        length = 1
        if varsizes is not None:
            if isinstance(varsizes, list):
                for size in varsizes:
                    if isinstance(size, dict):
                        for _, value in size.items():
                            assert isinstance(value, int)
                    else:
                        assert isinstance(size, int)
                length = len(varsizes)
            else:
                assert isinstance(varsizes, int)
                length = 1
        return length

    def set_early_stopping(self):
        """Enable early stopping."""
        self.early_stopping = True

    def unset_early_stopping(self):
        """Disable early stopping."""
        self.early_stopping = False

    def print_scaling_strategies(self):
        """ Print the list of the scaling strategies. """
        index = 0
        print("Scaling strategies:")
        for scaling in self.scalings:
            index += 1
            msg = "# " + str(index)
            for var_type, var_size in scaling.items():
                msg += " - "
                msg += var_type + ": " + str(var_size)
            print(msg)

    def execute(self, n_replicates=1):
        """Execute the scalability study, one or several times to take into
        account the random features of the scalable problems.

        :param int n_replicates: number of times the scalability study
            is repeated. Default: 1.
        """
        LOGGER.info("Execute scalability study %s times", n_replicates)
        if not self.formulations and not self.algorithms:
            raise ValueError(
                "A scalable study needs at least "
                "1 optimization strategy, "
                "defined by a mandatory optimization algorithm, "
                "and optional optimization algorithm, "
                "optimization options"
            )
        index = 0
        for algo, formulation in zip(self.algorithms, self.formulations):
            for id_scaling, scaling in enumerate(self.var_scalings):
                for replicate in range(1, n_replicates + 1):
                    problem = ScalableProblem(
                        self.datasets,
                        self.design_variables,
                        self.objective,
                        self.eq_constraints,
                        self.ineq_constraints,
                        self.maximize_objective,
                        scaling,
                        fill_factor=self._fill_factor,
                        seed=replicate,
                        group_dep=self._group_dep,
                        force_input_dependency=True,
                        allow_unused_inputs=False,
                    )
                    varnames = [algo, formulation, id_scaling + 1, replicate]
                    name = "_".join([self.prefix] + [str(var) for var in varnames])
                    if name[0] == "_":
                        name = name[1:]
                    directory = os.path.join(self.directory, POSTSCAL_DIRECTORY, name)
                    self.__create_directory(directory)
                    problem.plot_dependencies(True, False, directory)
                    form_opt = self.formulations_options[index]
                    if not isinstance(form_opt, dict):
                        formulation_options = {}
                    else:
                        formulation_options = form_opt
                    problem.create_scenario(
                        formulation,
                        "MDO",
                        self.start_at_equilibrium,
                        self.active_probability,
                        self.feasibility_level,
                        **formulation_options
                    )
                    top_level_disciplines = (
                        problem.scenario.formulation.get_top_level_disc()
                    )
                    for disc in top_level_disciplines:
                        disc.linearization_mode = self.top_level_diff[index]
                    algo_options = deepcopy(self.algorithms_options[index])
                    max_iter = algo_options["max_iter"]
                    del algo_options["max_iter"]
                    problem.scenario.execute(
                        {
                            "algo": algo,
                            "max_iter": max_iter,
                            "algo_options": algo_options,
                        }
                    )
                    stopidx, n_iter = self.__get_stop_index(problem)
                    ratio = float(stopidx) / n_iter
                    self.__save_opt_history_view(
                        problem,
                        algo + "_" + formulation,
                        str(id_scaling + 1),
                        str(replicate),
                    )
                    n_calls = {
                        disc: n_calls * ratio
                        for disc, n_calls in problem.n_calls.items()
                    }
                    tmp = problem.n_calls_linearize
                    n_calls_linearize = {disc: ncl * ratio for disc, ncl in tmp.items()}
                    tmp = problem.n_calls_top_level
                    n_calls_tl = {
                        disc: n_calls * ratio for disc, n_calls in tmp.items()
                    }
                    tmp = problem.n_calls_linearize_top_level
                    n_calls_linearize_tl = {
                        disc: ncltl * ratio for disc, ncltl in tmp.items()
                    }
                    exec_time = problem.exec_time()
                    exec_time *= ratio
                    status = problem.status
                    is_feasible = problem.is_feasible
                    result = ScalabilityResult(name, id_scaling + 1, replicate)
                    self.results.append(result)
                    inputs = problem.inputs
                    outputs = problem.outputs
                    disc_varnames = {
                        disc: inputs[disc] + outputs[disc]
                        for disc in problem.disciplines
                    }
                    sizes = problem.varsizes
                    new_varsizes = {
                        disc: {
                            name: scaling.get(name, sizes[name])
                            for name in disc_varnames[disc]
                        }
                        for disc in problem.disciplines
                    }
                    old_varsizes = problem.varsizes
                    result.get(
                        algo,
                        algo_options,
                        formulation,
                        formulation_options,
                        scaling,
                        n_calls,
                        n_calls_linearize,
                        n_calls_tl,
                        n_calls_linearize_tl,
                        exec_time,
                        status,
                        is_feasible,
                        problem.disciplines,
                        problem.outputs,
                        old_varsizes,
                        new_varsizes,
                    )
                    result.save(self.directory)
            index += 1
        return self.results

    def __save_opt_history_view(self, problem, optim, scaling, replicate):
        """Save optimization history view
        according to a directory tree structure.

        :param ScalableProblem problem: scalable problem
        :param str optim: optimization algorithm and MDO formulation
        :param str scaling: scaling index
        :param str replicate: replicate index
        """
        LOGGER.info("Save optimization history view")
        fpath = os.path.join(
            self.directory, POSTOPTIM_DIRECTORY, "_".join([self.prefix] + [optim])
        )
        self.__create_directory(fpath)
        fpath = os.path.join(fpath, "scaling_" + scaling)
        self.__create_directory(fpath)
        fpath = os.path.join(fpath, "replicate_" + replicate)
        self.__create_directory(fpath)
        fpath = fpath + "/"
        problem.scenario.post_process(
            "OptHistoryView", save=True, show=False, file_path=fpath
        )

    def __get_stop_index(self, problem):
        """Get stop index from a database.

        :param ScalableProblem problem: scalable problem
        :return: stop index, database length
        :rtype: int, int
        """
        database = problem.scenario.formulation.opt_problem.database
        n_iter = len(database)
        if self.early_stopping:
            y_prev = inf
            stopidx = 0
            for _, value in database.items():
                pbm = problem.scenario.formulation.opt_problem
                diff = abs(y_prev - value[pbm.get_objective_name()])
                diff /= abs(y_prev)
                if diff < 1e-6:
                    break
                y_prev = value[pbm.get_objective_name()]
                stopidx += 1
        else:
            stopidx = n_iter
        return stopidx, n_iter


class ScalabilityResult(object):
    """Scalability Result."""

    def __init__(self, name, id_scaling, id_sample):
        """Constructor.

        :param str name: name of the scalability result.
        :param int id_scaling: scaling identifiant
        :param int id_sample: sample identifiant
        """
        self.name = name
        self.id_scaling = id_scaling
        self.id_sample = id_sample
        self.algo = None
        self.algo_options = None
        self.formulation_options = None
        self.formulation = None
        self.scaling = None
        self.n_calls = None
        self.n_calls_linearize = None
        self.n_calls_top_level = None
        self.n_calls_linearize_top_level = None
        self.exec_time = None
        self.original_exec_time = None
        self.status = None
        self.is_feasible = None
        self.disc_names = None
        self.old_varsizes = None
        self.new_varsizes = None
        self.output_names = None

    def get(
        self,
        algo,
        algo_options,
        formulation,
        formulation_options,
        scaling,
        n_calls,
        n_calls_linearize,
        n_calls_top_level,
        n_calls_linearize_top_level,
        exec_time,
        status,
        is_feasible,
        disc_names,
        output_names,
        old_varsizes,
        new_varsizes,
    ):
        """Get a scalability result for a given optimization strategy
        and a given scaling strategy.

        :param str algo: name of the optimization algorithm
        :param dict algo_options: options of the optimization algorithm
        :param str formulation: name of the MDO formulation
        :param dict formulation_options: options of the MDO formulation
        :param scaling: scaling strategy
        :param list(int) n_calls: number of calls for each discipline
        :param list(int) n_calls_linearize: number of linearization
            for each discipline
        :param list(int) n_calls_top_level: number of calls for each discipline
        :param list(int) n_calls_linearize_top_level: number of linearization
            for each discipline
        :param float exec_time: execution time
        :param int status: status of the optimization scenario
        :param bool is_feasible: feasibility of the optimization solution
        :param list(str) disc_names: list of discipline names
        :param dict output_names: list of output names
        :param dict old_varsizes: old variable sizes
        :param dict new_varsizes: new variable sizes
        """
        self.algo = algo
        self.algo_options = algo_options
        self.formulation = formulation
        self.formulation_options = formulation_options
        self.scaling = scaling
        self.n_calls = n_calls
        self.n_calls_linearize = n_calls_linearize
        self.n_calls_top_level = n_calls_top_level
        self.n_calls_linearize_top_level = n_calls_linearize_top_level
        self.exec_time = exec_time
        self.status = status
        self.is_feasible = is_feasible
        self.disc_names = disc_names
        self.output_names = output_names
        self.old_varsizes = old_varsizes
        self.new_varsizes = new_varsizes

    def save(self, study_directory):
        """Save a scalability result into a pickle file
        whose name is the name of the ScalabilityResult instance."""
        fpath = os.path.join(study_directory, RESULTS_DIRECTORY, self.name + ".pkl")
        result = {
            "algo": self.algo,
            "algo_options": self.algo_options,
            "formulation": self.formulation,
            "formulation_options": self.formulation_options,
            "scaling": self.scaling,
            "n_calls": self.n_calls,
            "n_calls_linearize": self.n_calls_linearize,
            "n_calls_top_level": self.n_calls_top_level,
            "n_calls_linearize_top_level": self.n_calls_linearize_top_level,
            "exec_time": self.exec_time,
            "status": self.status,
            "is_feasible": self.is_feasible,
            "disc_names": self.disc_names,
            "output_names": self.output_names,
            "old_varsizes": self.old_varsizes,
            "new_varsizes": self.new_varsizes,
        }
        with open(fpath, "wb") as fout:
            pickle.dump(result, fout)

    def load(self, study_directory):
        """Load a scalability result from a pickle file
        whose name is the name of the ScalabilityResult instance."""
        fpath = os.path.join(study_directory, RESULTS_DIRECTORY, self.name + ".pkl")
        with open(fpath, "rb") as fin:
            result = pickle.load(fin)
        self.algo = result["algo"]
        self.formulation = result["formulation"]
        self.scaling = result["scaling"]
        self.n_calls = result["n_calls"]
        self.n_calls_linearize = result["n_calls_linearize"]
        self.n_calls_top_level = result["n_calls_top_level"]
        self.n_calls_linearize_top_level = result["n_calls_linearize_top_level"]
        self.exec_time = result["exec_time"]
        self.status = result["status"]
        self.is_feasible = result["is_feasible"]
        self.disc_names = result["disc_names"]
        self.output_names = result["output_names"]
        self.old_varsizes = result["old_varsizes"]
        self.new_varsizes = result["new_varsizes"]


class PostScalabilityStudy(object):

    """The PostScalabilityStudy class aims to post-process a list of
    scalability results stored in a directory.
    """

    NOMENCLATURE = {
        "exec_time": "Execution time (s)",
        "original_exec_time": "Pseudo-original execution time",
        "n_calls": "Number of discipline evaluations",
        "n_calls_linearize": "Number of gradient evaluations",
        "status": "Optimizatin status",
        "is_feasible": "Feasibility of the solution (0 or 1)",
        "scaling_strategy": "Scaling strategy index",
        "total_calls": "Total number of evaluations",
    }

    def __init__(self, study_directory):
        """Constructor.

        :param str study: directory of the scalability study."
        """
        LOGGER.info("*** Start post-processing for scalability study ***")
        LOGGER.info("Working directory: %s", study_directory)
        self.study_directory = study_directory
        self.scalability_results = self.__load_results()
        self.n_results = len(self.scalability_results)
        self.descriptions = self.NOMENCLATURE
        self.cost_function = {}
        self.unit_cost = None
        for result in self.scalability_results:
            result.total_calls = sum(result.n_calls.values())
            result.total_calls += sum(result.n_calls_linearize.values())

    def set_cost_function(self, formulation, cost):
        """Set cost function for each formulation.

        :param str formulation: name of the formulation.
        :param function cost: cost function
        """
        self.cost_function[formulation] = cost

    def set_cost_unit(self, cost_unit):
        """Set the measurement unit for cost evaluation.

        :param str cost_unit: cost unit, e.g. 'h', 'min', ...
        """
        self.unit_cost = cost_unit
        desc = self.descriptions["original_exec_time"]
        desc = desc.split(" (")[0]
        desc += " (" + cost_unit + ")"
        self.descriptions["original_exec_time"] = desc

    def labelize_exec_time(self, description):
        """Change the description of execution time,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("exec_time", description)

    def labelize_original_exec_time(self, description):
        """Change the description of original execution time,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("original_exec_time", description)
        if self.unit_cost is not None:
            self.set_cost_unit(self.unit_cost)

    def labelize_n_calls(self, description):
        """Change the description of number of calls,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("n_calls", description)

    def labelize_n_calls_linearize(self, description):
        """Change the description of number of calls for linearization,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("n_calls_linearize", description)

    def labelize_status(self, description):
        """Change the description of status,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("status", description)

    def labelize_is_feasible(self, description):
        """Change the description of feasibility,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("is_feasible", description)

    def labelize_scaling_strategy(self, description):
        """Change the description of scaling strategy,
        used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("scaling_strategy", description)

    def _update_descriptions(self, keyword, description):
        """Update the description initialized
        with the NOMENCLATURE class attribute.

        :param str keyword: keyword of the considered object.
        :param str description: new description
        """
        if not self.descriptions.get(keyword):
            raise ValueError(
                "The keyword " + keyword + " is not "
                "in the list: " + ", ".join(list(self.descriptions.keys()))
            )
        if not isinstance(description, basestring):
            raise TypeError(
                'The argument "description" must be '
                "of type string, "
                "not of type " + description.__class__.__name__
            )
        self.descriptions[keyword] = description

    def __load_results(self):
        """ Load results from the results directory of the study path."""
        if not os.path.isdir(self.study_directory):
            raise ValueError('Directory "' + self.study_directory + '" does not exist.')
        directory = os.path.join(self.study_directory, RESULTS_DIRECTORY)
        if not os.path.isdir(directory):
            raise ValueError('Directory "' + directory + '" does not exist.')
        filenames = [
            filename
            for filename in os.listdir(directory)
            if os.path.isfile(os.path.join(directory, filename))
        ]
        results = []
        for filename in filenames:
            name = filename.split(".")[0]
            id_scaling = int(name.split("_")[-2])
            id_sample = int(name.split("_")[-1])
            result = ScalabilityResult(name, id_scaling, id_sample)
            result.load(self.study_directory)
            results.append(result)
        if not results:
            raise ValueError("Directory " + directory + " is empty.")
        return results

    def plot(
        self,
        legend_loc="upper left",
        xticks=None,
        xticks_labels=None,
        xmargin=0.0,
        **options
    ):
        """Plot the results using different methods
        according to the presence or absence of replicate values.

        :param str legend_loc: legend localization
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) xticks_labels: list of xticks labels (default: None)
        :param float xmargin: margin on left and right sides of the x-axis
        :param options: options for the specialized plot methods
        """
        LOGGER.info("Execute post-processing")
        self.__has_scaling_dimension(xticks)
        self.__has_scaling_dimension(xticks_labels)
        there_are_replicates = len(self.get_replicates(True)) > 1
        if self.cost_function:
            self._estimate_original_time()
        if there_are_replicates:
            LOGGER.info("|    Type: replicate")
            self._replicate_plot(legend_loc, xticks, xticks_labels, xmargin, **options)
        else:
            LOGGER.info("|    Type: standard")
            self._standard_plot(legend_loc, xticks, xticks_labels, xmargin)

    def __has_scaling_dimension(self, value):
        """Assert if a value has the scaling dimension.

        :param ndarray value: value.
        """
        if value is not None and hasattr(value, "__len__"):
            assert len(value) == len(self.get_scaling_strategies(True))

    def _standard_plot(
        self, legend_loc="upper left", xticks=None, xticks_labels=None, xmargin=0.0
    ):
        """Deterministic plot.

        :param str legend_loc: legend localization
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) xticks_labels: list of xticks labels (default: None)
        :param float xmargin: margin on left and right sides of the x-axis
        """
        colors = ["blue", "red", "green"]
        idos = -1
        for optim_strategy in self.get_optimization_strategies(True):
            idos += 1
            color = colors[idos]
            xcoord, values = self.__get_values(optim_strategy)
            for ylabel, ycoord in values.items():
                if hasattr(ycoord, "shape"):
                    plt.figure(ylabel)
                    if len(ycoord.shape) == 3:
                        n_criteria = ycoord.shape[0]
                    else:
                        n_criteria = 1
                    linestyles = [
                        (0, (id_criterion + 1, n_criteria))
                        for id_criterion in range(n_criteria)
                    ]
                    for criterion in range(n_criteria):
                        if len(ycoord.shape) == 3:
                            yycoord = ycoord[criterion, :, 0].flatten()
                        else:
                            yycoord = ycoord[:, 0].flatten()
                        yval = yycoord
                        xval = list(xcoord)
                        if xticks is not None:
                            assert hasattr(xticks, "__len__")
                            assert len(xticks) == len(xval)
                            xval = xticks
                        else:
                            xticks = xval
                        plt.plot(
                            xval,
                            yval,
                            "o",
                            linestyle=linestyles[criterion],
                            color=color,
                        )
                        if xticks_labels is None:
                            xticks_labels = xticks
                        else:
                            xticks_labels = xval
                        plt.xticks(xticks, xticks_labels)
                        plt.xlabel(self.descriptions["scaling_strategy"])
                        plt.ylabel(self.descriptions[ylabel])
                        plt.xlim(xval[0] - xmargin, xval[-1] + xmargin)
        problem_legends = self.get_optimization_strategies(True)
        legend_elements = [
            Line2D([0], [0], color=colors[idx], lw=4, label=problem_legends[idx])
            for idx in range(len(problem_legends))
        ]
        for ylabel, ycoord in values.items():
            if hasattr(ycoord, "shape"):
                plt.figure(ylabel)
                plt.legend(
                    handles=legend_elements,
                    loc=legend_loc,
                    frameon=False,
                    framealpha=0.5,
                )
                fpath = os.path.join(
                    self.study_directory, POSTSTUDY_DIRECTORY, ylabel + ".pdf"
                )
                plt.savefig(fpath)

    @property
    def n_samples(self):
        """ Number of samples """
        return len(self.get_replicates(True))

    def _replicate_plot(
        self,
        legend_loc="upper left",
        xticks=None,
        xticks_labels=None,
        xmargin=0.0,
        minbox=2,
        notch=False,
        widths=0.25,
        whis=1.5,
    ):
        """Probabilistic plot.

        :param str legend_loc: legend localization
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) xticks_labels: list of xticks labels (default: None)
        :param float xmargin: margin on left and right sides of the x-axis
        :param int minbox: minimal number of values for boxplot (default: 2).
        :param bool notch: if True, will produced a notched boxplot.
        :param float whis:  the reach of the whiskers to the beyond
            the first and third quartiles (default: 1.5).
        """
        if not hasattr(widths, "__len__"):
            widths = [widths] * len(self.get_scaling_strategies(True))
        else:
            self.__has_scaling_dimension(widths)

        n_samples = len(self.get_replicates(True))
        colors = ["blue", "red", "green"]
        if n_samples >= minbox:
            idos = -1
            n_strategies = len(self.get_optimization_strategies(True))
            for optim_strategy in self.get_optimization_strategies(True):
                idos += 1
                xcoord, values = self.__get_values(optim_strategy)
                for ylabel, ycoord in values.items():
                    if hasattr(ycoord, "shape") and ylabel != "status":
                        plt.figure(ylabel)
                        xval = list(xcoord)
                        if xticks is not None:
                            assert hasattr(xticks, "__len__")
                            assert len(xticks) == len(xval)
                            xval = xticks
                        tmp = []
                        for idx, xtick in enumerate(xval):
                            tmp.append(float(idos) / n_strategies)
                            tmp[-1] *= widths[idx] * 3
                            tmp[-1] += xtick
                        self.__draw_boxplot(
                            ycoord, tmp, colors[idos], colors[idos], notch, widths, whis
                        )
                        if len(ycoord.shape) == 3:
                            zval = ycoord[0, :, :]
                        else:
                            zval = ycoord
                        xval_offset = [
                            xtick + float(idos) / n_strategies * widths[idx] * 3
                            for idx, xtick in enumerate(xval)
                        ]
                        plt.plot(xval_offset, median(zval, 1), "--", color=colors[idos])
                        if xticks_labels is None:
                            xticks_labels = xval
                        plt.xticks(xval, xticks_labels)
                        plt.xlim(xval[0] - xmargin, xval[-1] + xmargin)
                        plt.xlabel(self.descriptions["scaling_strategy"])
                        plt.ylabel(self.descriptions[ylabel])
                        plt.yscale("log")
                        plt.grid(True, "both")
        problem_legends = self.get_optimization_strategies(True)
        legend_elements = [
            Line2D([0], [0], color=colors[idx], lw=4, label=problem_legends[idx])
            for idx in range(len(problem_legends))
        ]
        for ylabel, ycoord in values.items():
            if hasattr(ycoord, "shape") and ylabel != "status":
                plt.figure(ylabel)
                plt.legend(
                    handles=legend_elements,
                    loc=legend_loc,
                    frameon=False,
                    framealpha=0.5,
                )
                fpath = os.path.join(
                    self.study_directory, POSTSTUDY_DIRECTORY, ylabel + ".pdf"
                )
                plt.savefig(fpath)

    @staticmethod
    def __draw_boxplot(data, xticks, edge_color, fill_color, notch, widths, whis):
        """Draw boxplot from a dataset.

        :param array data: dataset array of dimension 2 or 3
        :param list(float) xticks: values of xticks
        :param str edge_color: edge color
        :param str fill_color: fill color
        :param bool notch: if True, will produced a notched boxplot.
        :param list(float) widths: widths of boxplots
        :param float whis:  the reach of the whiskers to the beyond
            the first and third quartiles
        """
        if len(data.shape) == 3:
            data = data[0, :, :]
        boxplot = plt.boxplot(
            data.T,
            patch_artist=True,
            positions=xticks,
            showfliers=False,
            whiskerprops={"linestyle": "-"},
            notch=notch,
            widths=widths,
            whis=whis,
        )

        for element in ["boxes", "whiskers", "fliers", "means", "caps"]:
            plt.setp(boxplot[element], color=edge_color)
        plt.setp(boxplot["medians"], color="white")

        for patch in boxplot["boxes"]:
            patch.set(facecolor=fill_color)

        return boxplot

    def __get_values(self, optim_strategy):
        """Get values of criteria and corresponding scaling levels
        for a given optimization strategy and all replicates.

        :param str optim_strategy: name of the optimization strategy.
        :return: scaling levels, criteria values
        :rtype: list(int), dict(array)
        """
        exec_time = []
        original_exec_time = []
        n_calls = []
        n_calls_linearize = []
        status = []
        is_feasible = []
        scaling = []
        total_calls = []
        for replicate in self.get_replicates(True):
            xcoord, ycoord = self.__get_replicate_values(optim_strategy, replicate)
            exec_time.append(ycoord["exec_time"])
            total_calls.append(ycoord["total_calls"])
            if self.cost_function:
                original_exec_time.append(ycoord["original_exec_time"])
            tmp = [
                [val for _, val in ycoord["n_calls"][idx].items()]
                for idx in range(len(ycoord["n_calls"]))
            ]
            n_calls.append(tmp)
            tmp = [
                [val for _, val in ycoord["n_calls_linearize"][idx].items()]
                for idx in range(len(ycoord["n_calls_linearize"]))
            ]
            n_calls_linearize.append(tmp)
            status.append(ycoord["status"])
            is_feasible.append(ycoord["is_feasible"])
            scaling.append(ycoord["scaling"])
        scaling_levels = xcoord
        values = {}
        values["exec_time"] = array(exec_time).T
        if self.cost_function:
            values["original_exec_time"] = array(original_exec_time).T
        values["n_calls"] = array(n_calls).T
        values["n_calls_linearize"] = array(n_calls_linearize).T
        values["total_calls"] = array(total_calls).T
        values["status"] = array(status).T
        values["is_feasible"] = array(is_feasible).T
        values["scaling"] = scaling
        return scaling_levels, values

    def __get_replicate_values(self, optim_strategy, replicate):
        """Get values of criteria and corresponding scaling levels
        for a given optimization strategy and a given replicate.

        :param str optim_strategy: optimization strategy.
        :param int replicate: replicate index.
        :return: scaling levels, criteria values
        :rtype: list(int), dict(array-like)
        """
        are_replicate = [value == replicate for value in self.get_replicates()]
        are_optim_strategy = [
            value == optim_strategy for value in self.get_optimization_strategies()
        ]
        are_ok = [
            is_rep and is_oo for is_rep, is_oo in zip(are_replicate, are_optim_strategy)
        ]
        indices = [index for index, is_ok in enumerate(are_ok) if is_ok]
        scaling_levels = self.get_scaling_strategies()
        scaling_levels = [scaling_levels[index] for index in indices]
        tmp = sorted(list(range(len(scaling_levels))), key=lambda k: scaling_levels[k])
        indices = [indices[index] for index in tmp]
        scaling_levels = [scaling_levels[index] for index in tmp]
        results = [self.scalability_results[index] for index in indices]
        exec_time = [result.exec_time for result in results]
        total_calls = [result.total_calls for result in results]
        original_exec_time = [result.original_exec_time for result in results]
        n_calls = [result.n_calls for result in results]
        n_calls_linearize = [result.n_calls_linearize for result in results]
        status = [result.status for result in results]
        is_feasible = [result.is_feasible for result in results]
        scaling = [result.scaling for result in results]
        values = {}
        values["exec_time"] = exec_time
        values["total_calls"] = total_calls
        if self.cost_function:
            values["original_exec_time"] = original_exec_time
        values["n_calls"] = n_calls
        values["n_calls_linearize"] = n_calls_linearize
        values["status"] = status
        values["is_feasible"] = is_feasible
        values["scaling"] = scaling
        return scaling_levels, values

    @property
    def names(self):
        """ Get the names of the scalability results. """
        return [value.name for value in self.scalability_results]

    def get_optimization_strategies(self, unique=False):
        """Get the names of the optimization strategies.

        :param bool unique: return either unique values if True
            or one value per scalability result if False (default: False).
        :return: names of the optimization stategies.
        :rtype: list(str)
        """
        os_list = ["_".join(name.split("_")[0:-2]) for name in self.names]
        if unique:
            os_list = sorted(set(os_list))
        return os_list

    def get_scaling_strategies(self, unique=False):
        """Get the identificants of the scaling strategies.

        :param bool unique: return either unique values if True
            or one value per scalability result if False (default: False).
        :return: identifiants of scaling strategies
        :rtype: list(int)
        """
        ss_list = [int(name.split("_")[-2]) for name in self.names]
        if unique:
            ss_list = sorted(set(ss_list))
        return ss_list

    def get_replicates(self, unique=False):
        """Get the replicate identifiants.

        :param bool unique: return either unique values if True
            or one value per scalability result if False (default: False).
        :return: identifiants of replicates.
        :rtype: list(int)
        """
        rep = [int(name.split("_")[-1]) for name in self.names]
        if unique:
            rep = sorted(set(rep))
        return rep

    def _estimate_original_time(self):
        """Estimate the original execution time from the number of calls
        and linearizations of the different disciplines and
        top-level disciplines and from the cost functions provided by the
        user.

        :return: original time
        :rtype: float
        """
        for scalability_result in self.scalability_results:
            n_c = scalability_result.n_calls
            n_cl = scalability_result.n_calls_linearize
            n_tl_c = scalability_result.n_calls_top_level
            n_tl_cl = scalability_result.n_calls_linearize_top_level
            varsizes = scalability_result.new_varsizes
            formulation = scalability_result.formulation
            if formulation not in self.cost_function:
                raise ValueError(
                    "The cost function of "
                    + formulation
                    + " must be defined in order to compute "
                    " the estimated original time."
                )
            result = self.cost_function[formulation](
                varsizes, n_c, n_cl, n_tl_c, n_tl_cl
            )
            scalability_result.original_exec_time = result
